import unittest
from Atomic import diff
import copy


class FakeObj(object):
    def __init__(self):
        self.name = None
        self.inspect_data = None
        self.metadata_results = None


class TestMetaDiff(unittest.TestCase):

    ORANGE1 = ['navel', 'caracara']
    ORANGE2 = ORANGE1 + ['valencia']

    APPLES1 = {
        'red': ['fuji', 'gala', 'delicious'],
        'green': ['granny', 'golden']
    }

    APPLES2 = copy.deepcopy(APPLES1)
    APPLES2['red'] = APPLES1['red'] + ['jazz']
    APPLES2['green'] = []

    MARKET = {
        'apples': {
            'gala': 1965,
            'fuji': 1962,
            'delicious': 1914
        }
    }

    MARKET2 = {
        'apples': {
            'gala': 1965,
            'honeycrisp': 1991,
            'fuji': 1962
        }
    }

    FRUIT1= {
        "a": "apple",
        "b": "banana",
        "c": "carrot",
        "d": ["dragonfruit", "dates", "dill"],
        "e": {"egg": "EGG",
              "endive": 'ENDIVE'},
        "f": "figs"
    }

    FRUIT2 = copy.deepcopy(FRUIT1)
    FRUIT2['g'] = 'grapes'

    FRUIT3 = copy.deepcopy(FRUIT1)
    FRUIT3['o'] = ORANGE1
    FRUIT4 = copy.deepcopy(FRUIT1)
    FRUIT4['o'] = ORANGE2

    FRUIT5 = copy.deepcopy(FRUIT1)
    FRUIT5['o'] = APPLES1
    FRUIT6 = copy.deepcopy(FRUIT1)
    FRUIT6['o'] = APPLES2

    FRUIT7 = copy.deepcopy(FRUIT1)
    FRUIT7['r'] = MARKET
    FRUIT8 = copy.deepcopy(FRUIT1)
    FRUIT8['r'] = MARKET2

    EMPTY1 = {'foo': None}
    EMPTY2 = {}

    NEST_NONE1 = {'foo': {'bar': {'baz': 1}}}
    NEST_NONE2 = {'foo': {'bar': None}}

    def _create_image_list(self, *images):
        image_list = []
        for i in images:
            _image = FakeObj()
            _image.name = i
            _image.inspect_data = getattr(self, i)
            image_list.append(_image)
        return image_list

    @staticmethod
    def _get_results_metadata(images):
        return images[0].metadata_results, images[1].metadata_results

    def test_same(self):
        images = self._create_image_list('FRUIT1', 'FRUIT1')
        compare = diff.CompareMetaData(images)
        compare.compare()
        self.assertEqual(compare.img_obj1.metadata_results,
                         compare.img_obj2.metadata_results)

    def test_different_single(self):
        images = self._create_image_list('FRUIT1', 'FRUIT2')
        compare = diff.CompareMetaData(images)
        compare.compare()
        results1, results2 = self._get_results_metadata(images)
        self.assertEqual(results1, {})
        self.assertEqual(results2, {'g': 'grapes'})

    def test_list_same(self):
        images = self._create_image_list('FRUIT3', 'FRUIT3')
        compare = diff.CompareMetaData(images)
        compare.compare()
        results1, results2 = self._get_results_metadata(images)
        self.assertEqual(results1, results2)
        self.assertEqual(results1, {})

    def test_list_different(self):
        images = self._create_image_list('FRUIT3', 'FRUIT4')
        compare = diff.CompareMetaData(images)
        compare.compare()
        results1, results2 = self._get_results_metadata(images)
        self.assertEqual(results1, {'o': ['navel', 'caracara']} )
        self.assertEqual(results2, {'o': ['navel', 'caracara', 'valencia']})

    def test_nested_dict_same(self):
        images = self._create_image_list('FRUIT5', 'FRUIT5')
        compare = diff.CompareMetaData(images)
        compare.compare()
        results1, results2 = self._get_results_metadata(images)
        self.assertEqual(results1, results2)
        self.assertEqual(results1, {})

    def test_nested_dict_diff(self):
        images = self._create_image_list('FRUIT5', 'FRUIT6')
        compare = diff.CompareMetaData(images)
        compare.compare()
        results1, results2 = self._get_results_metadata(images)
        self.assertEqual(results1, {'o': {'green': ['granny', 'golden'], 'red': ['fuji', 'gala', 'delicious']}} )
        self.assertEqual(results2, {'o': {'green': [], 'red': ['fuji', 'gala', 'delicious', 'jazz']}})

    def test_nested2_dict_same(self):
        images = self._create_image_list('FRUIT7', 'FRUIT7')
        compare = diff.CompareMetaData(images)
        compare.compare()
        results1, results2 = self._get_results_metadata(images)
        self.assertEqual(results1, results2)
        self.assertEqual(results1, {})

    def test_nested2_dict_diff(self):
        images = self._create_image_list('FRUIT7', 'FRUIT8')
        compare = diff.CompareMetaData(images)
        compare.compare()
        results1, results2 = self._get_results_metadata(images)
        self.assertEqual(results1, {'r': {'apples': {'delicious': 1914}}})
        self.assertEqual(results2, {'r': {'apples': {'honeycrisp': 1991}}})

    def test_none_to_empty(self):
        images = self._create_image_list('EMPTY1', 'EMPTY2')
        compare = diff.CompareMetaData(images)
        compare.compare()
        results1, results2 = self._get_results_metadata(images)
        self.assertNotEqual(results1, results2)
        self.assertEqual(results2, {})
        self.assertEqual(results1, {'foo': None})

    def test_nested_none(self):
        images = self._create_image_list('NEST_NONE1', 'NEST_NONE2')
        compare = diff.CompareMetaData(images)
        compare.compare()
        results1, results2 = self._get_results_metadata(images)
        self.assertEqual(results1, {'foo': {'bar': {'baz': 1}}})
        self.assertEqual(results2, {'foo': {'bar': None}})



from . import Atomic
from . import util
from Atomic.backendutils import BackendUtils

ATOMIC_CONFIG = util.get_atomic_config()
storage = ATOMIC_CONFIG.get('default_storage', "docker")

class Tag(Atomic):
    def __init__(self):
        super(Tag, self).__init__()
        self.be = None

    def tag_image(self):
        """
        Tag an image with a different name
        :return: 0 if the tag was created
        """
        if self.args.debug:
            util.write_out(str(self.args))

        beu = BackendUtils()

        backend = None
        if self.args.storage:
            backend = beu.get_backend_from_string(self.args.storage)
            image  = backend.has_image(self.args.src)

        else:            
            backend, image = beu.get_backend_and_image_obj(self.args.src, required=False)

        if not backend or not image:
            raise ValueError("Cannot find image {}.".format(self.args.src))

        backend.tag_image(self.args.src, self.args.target)

        # We need to return something here for dbus
        return 0



import argparse
import re
import sys
from . import util
from .util import add_opt
from .syscontainers import OSTREE_PRESENT
from Atomic.backendutils import BackendUtils
from Atomic.discovery import RegistryInspectError
from time import gmtime, strftime

try:
    from . import Atomic
except ImportError:
    from atomic import Atomic # pylint: disable=relative-import

INSTALL_ARGS = ["run",
                "-t",
                "-i",
                "--rm",
                "--privileged",
                "-v", "/:/host",
                "--net=host",
                "--ipc=host",
                "--pid=host",
                "-e", "HOST=/host",
                "-e", "NAME=${NAME}",
                "-e", "IMAGE=${IMAGE}",
                "-e", "CONFDIR=/host/etc/${NAME}",
                "-e", "LOGDIR=/host/var/log/${NAME}",
                "-e", "DATADIR=/host/var/lib/${NAME}",
                "-e", "SYSTEMD_IGNORE_CHROOT=1", 
                "--name", "${NAME}",
                "${IMAGE}"]

ATOMIC_CONFIG = util.get_atomic_config()
_storage = ATOMIC_CONFIG.get('default_storage', "docker")

def cli(subparser):
    # atomic install
    installp = subparser.add_parser(
        "install", help=_("execute container image install method"),
        epilog="atomic install attempts to read the LABEL INSTALL field "
        "in the image, if it does not exist atomic will just pull "
        "the image on to your machine.  You could add a LABEL "
        "INSTALL command to your Dockerfile like: 'LABEL INSTALL "
        "%s'" % Install.print_install())
    installp.set_defaults(_class=Install, func='install')
    add_opt(installp)
    installp.add_argument("--storage", dest="storage", default=None,
                   help=_("Specify the storage. Default is currently '%s'.  You can"
                          " change the default by editing /etc/atomic.conf and changing"
                          " the 'default_storage' field." % _storage))
    installp.add_argument("-n", "--name", dest="name", default=None,
                          help=_("name of container"))
    installp.add_argument(
        "--display",
        default=False,
        action="store_true",
        help=_("preview the command that %s would execute") % sys.argv[0])
    installp.add_argument("image", help=_("container image"))
    if OSTREE_PRESENT:
        system_xor_user = installp.add_mutually_exclusive_group()
        system_xor_user.add_argument("--user", dest="user", action="store_true", default=False,
                                     help=_("install the image as an user image."))
        system_xor_user.add_argument("--system", dest="system",
                                     action='store_true', default=False,
                                     help=_('install a system container'))
        installp.add_argument("--system-package", dest="system_package", default="auto",
                              help=_('control how to install the package.  It accepts `auto`, `yes`, `no`, `build`'))
        installp.add_argument("--rootfs", dest="remote",
                              help=_("choose an existing exploded container/image to use "
                                     "its rootfs as a remote, read-only rootfs for the "
                                     "container to be installed"))
        installp.add_argument("--set", dest="setvalues",
                              action='append',
                              help=_("specify a variable in the VARIABLE=VALUE "
                                     "form for a system container"))
    installp.add_argument("args", nargs=argparse.REMAINDER,
                          help=_("additional arguments appended to the image "
                                 "install method"))


class Install(Atomic):
    def install(self):
        if self.args.debug:
            util.write_out(str(self.args))
        storage_set = False if self.args.storage is None else True
        storage = _storage if not storage_set else self.args.storage
        args_system = getattr(self.args, 'system', None)
        args_user= getattr(self.args, 'user', None)
        if (args_system or args_user) and storage != 'ostree' and storage_set:
            raise ValueError("The --system and --user options are only available for the 'ostree' storage.")
        be_utils = BackendUtils()
        try:
            # Check to see if the container already exists
            _, _ = be_utils.get_backend_and_container_obj(self.name)
            raise ValueError("A container '%s' is already present" % self.name)
        except ValueError:
            pass

        if self.user:
            if not util.is_user_mode():
                raise ValueError("--user does not work for privileged user")
            return self.syscontainers.install_user_container(self.image, self.name)
        if self.ostree_uri(self.image):
            return self.syscontainers.install(self.image, self.name)
        # Check if image exists
        str_backend = 'ostree' if self.args.system else self.args.storage or storage
        be = be_utils.get_backend_from_string(str_backend)
        img_obj = be.has_image(self.args.image)
        if img_obj and img_obj.is_system_type:
            be = be_utils.get_backend_from_string('ostree')

        if img_obj is None:
            # Unable to find the image locally, look remotely
            remote_image_obj = be.make_remote_image(self.args.image)
            # We found an atomic.type of system, therefore install it onto the ostree
            # backend
            if remote_image_obj.is_system_type and not storage_set:
                be_utils.message_backend_change('docker', 'ostree')
                be = be_utils.get_backend_from_string('ostree')
            be.pull_image(self.args.image, remote_image_obj, debug=self.args.debug)
            img_obj = be.has_image(self.image)

        if be.backend is not 'docker':
            if OSTREE_PRESENT and self.args.setvalues and not self.user and not self.system:
                raise ValueError("--set is valid only when used with --system or --user")

            # We need to fix this long term and get ostree
            # using the backend approach vs the atomic args
            be.syscontainers.set_args(self.args)
            return be.install(self.image, self.name)

        installation = None
        if storage == 'docker' and not args_system:
            if self.args.system_package == 'build':
                raise ValueError("'--system-package=build' is not supported for docker backend")
            installation = be.rpm_install(img_obj, self.name)

        install_args = img_obj.get_label('INSTALL')

        if installation or install_args:
            try:
                name = img_obj.fq_name
            except RegistryInspectError:
                name = img_obj.input_name
            install_data_content = {
                'id': img_obj.id,
                "container_name": self.name,
                'install_date': strftime("%Y-%m-%d %H:%M:%S", gmtime())
            }
            if installation:
                # let's fail the installation if rpm for this image is already installed
                if util.InstallData.image_installed(img_obj):
                    raise ValueError("Image {} is already installed.".format(self.image))
                install_data_content["rpm_installed_files"] = installation.installed_files
                rpm_nvra = re.sub(r"\.rpm$", "", installation.original_rpm_name)
                install_data_content["system_package_nvra"] = rpm_nvra
            install_data = {name: install_data_content}

        if not install_args:
            return 0
        install_args = install_args.split()
        cmd = self.sub_env_strings(self.gen_cmd(install_args + self.quote(self.args.args)))
        self.display(cmd)

        if not self.args.display:
            result = util.check_call(cmd)
            if result == 0:
                if installation or install_args:
                    # Only write the install data if the installation worked.
                    util.InstallData.write_install_data(install_data)
            return result



    @staticmethod
    def print_install():
        return "%s %s %s" % (util.default_docker(), " ".join(INSTALL_ARGS), "/usr/bin/INSTALLCMD")

    @staticmethod
    def ostree_uri(image_name):
        for i in ['dockertar:', 'ostree:', 'docker:']:
            if image_name.startswith(i):
                return True
        return False



#
# Project Atomic Unit Test Example Template
#
#pylint: skip-file
import unittest


class TestAtomicUnit(unittest.TestCase):
    """
    TODO: This test case ...
    """

    def test_unit_desired_behavior(self):
        """
        TODO: Verifies ...
        """
        self.assertTrue(True)


if __name__ == '__main__':
    unittest.main()



import argparse
import sys
from Atomic.backendutils import BackendUtils
from Atomic.backends._docker import DockerBackend
from . import util
from Atomic.discovery import RegistryInspectError

try:
    from . import Atomic
except ImportError:
    from atomic import Atomic # pylint: disable=relative-import

SPC_ARGS = ["-i",
            "--privileged",
            "-v", "/:/host",
            "-v", "/run:/run",
            "-v", "/etc/localtime:/etc/localtime",
            "-v", "/sys/fs/selinux:/sys/fs/selinux:ro",
            "--net=host",
            "--ipc=host",
            "--pid=host",
            "-e", "HOST=/host",
            "-e", "NAME=${NAME}",
            "-e", "IMAGE=${IMAGE}",
            "-e", "SYSTEMD_IGNORE_CHROOT=1",
            "--name", "${NAME}",
            "${IMAGE}"]

RUN_ARGS = ["-i",
            "--name", "${NAME}",
            "${IMAGE}"]

ATOMIC_CONFIG = util.get_atomic_config()
_storage = ATOMIC_CONFIG.get('default_storage', "docker")


def cli(subparser):
    # atomic run
    runp = subparser.add_parser(
        "run", help=_("execute container image run method"),
        epilog="atomic run attempts to start an existing container if a container "
        "name is specified, or execute container image run method if an image "
        "name is specified.  Defaults to the following command, if image does "
        "not specify LABEL run:\n'%s'" % Run.print_run())
    runp.set_defaults(_class=Run, func='run')
    run_group = runp.add_mutually_exclusive_group()
    util.add_opt(runp)
    runp.add_argument("--replace", "-r", dest="replace", default=False,
                      action="store_true", help=_("Replaces an existing container by the same name"
                                                  "if it exists."))
    runp.add_argument("--storage", dest="storage", default=None,
                          help=_("Specify the storage. Default is currently '%s'.  You can"
                                 " change the default by editing /etc/atomic.conf and changing"
                                 " the 'default_storage' field." % _storage))
    runp.add_argument("-n", "--name", dest="name", default=None,
                      help=_("name of container"))
    runp.add_argument("--spc", default=False, action="store_true",
                      help=_("use super privileged container mode: '%s'" %
                             Run.print_spc()))
    runp.add_argument("-d", "--detach", default=False, action="store_true",
                      help=_("run the container in the background"))
    runp.add_argument("image", help=_("container image"))
    runp.add_argument("command", nargs=argparse.REMAINDER,
                      help=_("command to execute within the container. "
                             "If container is not running, command is appended "
                             "to the image run method"))
    run_group.add_argument("--quiet", "-q", action="store_true",
                      help=_("Be less verbose."))

    run_group.add_argument(
        "--display",
        default=False,
        action="store_true",
        help=_("preview the command that %s would execute") % sys.argv[0])


class Run(Atomic):
    def __init__(self):
        super(Run, self).__init__()
        self.RUN_ARGS = RUN_ARGS
        self.SPC_ARGS = SPC_ARGS

    def run(self):
        storage_set = False if self.args.storage is None else True
        storage = _storage if not storage_set else self.args.storage
        be_utils = BackendUtils()
        if self.name:
            try:
                be, con_obj = be_utils.get_backend_and_container_obj(self.name)
                return be.run(con_obj, atomic=self, args=self.args)
            except ValueError:
                pass
        be = be_utils.get_backend_from_string(storage)
        db = DockerBackend()
        img_object = be.has_image(self.image)
        if img_object is None and storage == 'docker':
            self.display("Need to pull %s" % self.image)
            remote_image_obj = db.make_remote_image(self.args.image)
            # If the image has a atomic.type of system, then we need to land
            # this in the ostree backend.  Install it and then start it
            # because this is run
            if remote_image_obj.is_system_type and not storage_set:
                be = be_utils.get_backend_from_string('ostree')
                be_utils.message_backend_change('docker', 'ostree')
                be.install(self.image, self.name)
                con_obj = be.has_container(self.name)
                return be.run(con_obj)
            if self.args.display:
                return 0
            try:
                db.pull_image(self.image, remote_image_obj)
                img_object = db.has_image(self.image)
            except RegistryInspectError:
                raise ValueError("Unable to find image {}".format(self.image))
        return be.run(img_object, atomic=self, args=self.args)

    @staticmethod
    def print_run():
        return "%s run %s" % (util.default_docker(), " ".join(RUN_ARGS))

    @staticmethod
    def print_spc():
        return "%s run %s" % (util.default_docker(), " ".join(SPC_ARGS))



from . import util
from . import Atomic
from . import discovery
import os
import tempfile

try:
    from urlparse import urlparse #pylint: disable=import-error
except ImportError:
    from urllib.parse import urlparse #pylint: disable=no-name-in-module,import-error


ATOMIC_CONFIG = util.get_atomic_config()
READ_URIS = ['file', 'http', 'https']
WRITE_URIS = ['file']

def cli(subparser):
    # atomic sign
    signer = util.get_atomic_config_item(['default_signer'])
    gnupghome = util.getgnuhome()

    signp = subparser.add_parser("sign",
                                 help="Sign an image",
                                 epilog="Create a signature for an image which can be "
                                        "used later to verify it.")
    signp.set_defaults(_class=Sign, func="sign")
    signp.add_argument("images", nargs="*", help=_("images to sign"))
    signp.add_argument("--sign-by", dest="sign_by", default=signer,
                       help=_("Name of the signing key. Currently %s, "
                              "default can be defined in /etc/atomic.conf" % signer))
    signp.add_argument("-d", "--directory",
                       default=None,
                       dest="signature_path",
                       help=_("Define an alternate directory to store signatures"))
    signp.add_argument("-g", "--gnupghome",
                       default=gnupghome,
                       dest="gnupghome",
                       help=_("Set the GNUPGHOME environment variable to "
                              "use an alternate user's GPG keyring. "
                              "Useful when running with sudo, "
                              "e.g. set to '~/.gnupg'. "
                              "Default is %s" % gnupghome
                       ))

class Sign(Atomic):
    def __init__(self): # pylint: disable=useless-super-delegation
        super(Sign, self).__init__()

    def sign(self, in_signature_path=None, images=None):
        def no_reg_no_default_error(image, registry_path):
            return "Unable to associate {} with configurations in {} and " \
                   "no 'default_store' is defined.".format(image,
                                                           registry_path)

        if in_signature_path is None and getattr(self.args, 'signature_path', None) is not None:
            in_signature_path = self.args.signature_path

        if images is None:
            images = self.args.images

        if self.args.debug:
            util.write_out(str(self.args))

        signer = self.args.sign_by

        if self.args.sign_by is None:
            raise ValueError("No default identity (default_signer) was defined in /etc/atomic.conf "
                             "and no --sign-by identity was provided.  You must provide an identity")
        registry_config_path = util.get_atomic_config_item(["registry_confdir"], ATOMIC_CONFIG, '/etc/containers/registries.d')
        registry_configs, default_store = util.get_registry_configs(registry_config_path)

        # we honor GNUPGHOME if set, override with atomic.conf, arg overrides all
        if self.args.gnupghome:
            os.environ['GNUPGHOME'] = self.args.gnupghome

        for sign_image in images:
            registry, repo, image, tag, _ = util.Decompose(sign_image).all
            ri = discovery.RegistryInspect(registry, repo, image, tag, debug=self.args.debug, orig_input=sign_image)
            manifest = ri.get_manifest(return_json=False)


            try:
                manifest_file = tempfile.NamedTemporaryFile(mode="wb", delete=False)
                manifest_file.write(manifest)
                manifest_file.close()
                manifest_hash = str(util.skopeo_manifest_digest(manifest_file.name))
                expanded_image_name = ri.fqdn
                expanded_image_name_components = util.Decompose(expanded_image_name)

                if in_signature_path:
                    if not os.path.exists(in_signature_path):
                        raise ValueError("The path {} does not exist".format(in_signature_path))
                    signature_path = in_signature_path

                else:
                    config_key = "{}/{}".format(expanded_image_name_components.registry,
                                                expanded_image_name_components.repo)
                    if not registry_configs and not default_store:
                        raise ValueError(no_reg_no_default_error(sign_image, registry_config_path))
                    reg_info = util.have_match_registry(config_key, registry_configs)
                    if not reg_info:
                        reg_info = default_store
                        if reg_info is None:
                            raise ValueError("No applicable configuration for {} was found in {}".format(config_key, registry_config_path))

                    signature_path = util.get_signature_write_path(reg_info)
                    if signature_path is None:
                        raise ValueError("No write path for {} was "
                                         "found in {}".format(config_key, registry_config_path))
                    elif urlparse(signature_path).scheme not in WRITE_URIS:
                        raise ValueError("Writing to {} is not supported. Use a supported scheme {} "
                                         "instead.".format(urlparse(signature_path).scheme, WRITE_URIS))

                    # Deal with write path prepends
                    if urlparse(signature_path).scheme in WRITE_URIS:
                        signature_path = urlparse(signature_path).path

                    # Make sure signature path exists
                    if not os.path.exists(signature_path):
                        raise ValueError("The signature path {} does not exist".format(signature_path))

                # remote_path contains neither the registry hostname nor a digest/tag
                if expanded_image_name_components.repo:
                    remote_path = expanded_image_name_components.repo + '/' + expanded_image_name_components.image
                else:
                    remote_path = expanded_image_name_components.image
                sigstore_path = "{}/{}@{}".format(signature_path, remote_path, manifest_hash.replace(':', '=', 1))

                self.make_sig_dirs(sigstore_path)
                sig_name = self.get_sig_name(sigstore_path)
                fq_sig_path = os.path.join(sigstore_path, sig_name)
                if os.path.exists(fq_sig_path):
                    raise ValueError("The signature {} already exists.  If you wish to "
                                     "overwrite it, please delete this file first")

                util.skopeo_standalone_sign(expanded_image_name, manifest_file.name,
                                            self.get_fingerprint(signer), fq_sig_path, debug=self.args.debug)
                util.write_out("Created: {}".format(fq_sig_path))

            finally:
                os.remove(manifest_file.name)

    @staticmethod
    def get_fingerprint(signer):
        cmd = ['gpg2', '--no-permission-warning', '--with-colons', '--fingerprint', signer]
        stdout = util.check_output(cmd)
        for line in stdout.splitlines():
            _line = line.decode('utf-8')
            if _line.startswith('fpr:'):
                return _line.split(":")[9]

    @staticmethod
    def make_sig_dirs(sig_path):
        if not os.path.exists(sig_path):
            # TODO # pylint: disable=fixme
            # perhaps revisit directory permissions
            # when complete use-cases are known
            os.makedirs(sig_path)

    @staticmethod
    def get_sig_name(sig_path):
        sig_files = set(os.listdir(sig_path))
        sig_int = 1
        while True:
            name = "signature-{}".format(sig_int)
            if name not in sig_files:
                return name
            sig_int += 1



import os
import shutil
import tempfile

from docker import errors

import Atomic.util as util
from Atomic.backends.backend import Backend
from Atomic.client import AtomicDocker, no_shaw
from Atomic.objects.image import Image
from Atomic.objects.container import Container
from requests import exceptions
from Atomic.rpm_host_install import RPMHostInstall
from Atomic.trust import Trust
from Atomic.objects.layer import Layer
from dateutil.parser import parse as dateparse
from Atomic import Atomic
from requests.exceptions import HTTPError
from Atomic.backends._docker_errors import NoDockerDaemon
from Atomic.discovery import RegistryInspectError
from Atomic.atomic import AtomicError
from subprocess import CalledProcessError

try:
    from subprocess import DEVNULL  # pylint: disable=no-name-in-module
except ImportError:
    DEVNULL = open(os.devnull, 'wb')


class ContainerInstallation(object):
    """
    Provides variables which hold data for how build and
    installation process of a container went
    """
    def __init__(self, original_rpm_name, destination_path, installed_files):
        """
        :param original_rpm_name: verbose RPM name
        :param destination_path: path to built RPM
        :param installed_files: list of files provided by the built RPM
        """
        self.original_rpm_name = original_rpm_name
        self.destination_path = destination_path
        self.installed_files = installed_files


def build_rpm_for_docker_backend(image, name, temp_dir, labels):
    """
    build rpm package for specified docker image

    :param image, instance of Atomic.objects.image.Image
    :param name, str, name of the associated container
    :param temp_dir: str, directory where all the data will be processed
    :param labels: dict, these labels come from container image
    :return: instance of StandaloneContainerInstallation
    """
    from Atomic.mount import DockerMount, MountContextManager
    mount_path = os.path.join(temp_dir, "mountpoint")
    destination = os.path.join(temp_dir, "system_rpm")
    os.makedirs(destination)
    os.makedirs(mount_path)
    dm = DockerMount(mount_path)
    cm = MountContextManager(dm, image.id)
    with cm:
        # if we are on devicemapper, the path to container is <mount_point>/hostfs/
        dm_candidate_path = os.path.join(cm.mnt_path, "rootfs")
        if os.path.exists(dm_candidate_path):
            exports_dir = os.path.join(dm_candidate_path, "exports")
        else:
            exports_dir = os.path.join(cm.mnt_path, "exports")
        r = RPMHostInstall.generate_rpm(
            name, image.id, labels, exports_dir, destination)
        return ContainerInstallation(r[0], r[1], r[2])


class DockerBackend(Backend):
    def __init__(self):
        self.input = None
        self._d = None

    @property
    def available(self):
        try:
            _ = self.d
            return True
        except NoDockerDaemon:
            return False

    @property
    def d(self):
        if not self._d:
            self._d = AtomicDocker()
            self._ping()
            return self._d
        return self._d

    @property
    def backend(self):
        return "docker"

    def has_image(self, img):
        err_append = "Refine your search to narrow results."
        self.input = img
        image_info = self._get_images(get_all=True)

        img_obj = self.inspect_image(image=img)
        if img_obj:
            return img_obj
        # If we cannot find the image locally AND it has a digest
        # then bail.
        if '@sha256:' in img:
            return None
        name_search = util.image_by_name(img, images=image_info)
        length = len(name_search)
        if length == 0:
            # No dice
            return None
        if length > 1:
            tmp_image = dict((x['Id'], x['RepoTags']) for x in image_info)
            repo_tags = []
            for name in name_search:
                for repo_tag in tmp_image.get(name['Id']):
                    if repo_tag.find(img) > -1:
                        repo_tags.append(repo_tag)
            raise ValueError("Found more than one image possibly "
                             "matching '{0}'. They are:\n    {1} \n{2}"
                             .format(img, "\n    ".join(repo_tags),
                                     err_append))

        inspected = self._inspect_image(img)
        if not inspected:
            return None
        return self._make_image(img, inspected, deep=True)

    def already_has_image(self, local_img, remote_img):
        """
        Returns bool response if the image is already present.  Input must be an image object
        :param img_obj: an image object
        :return:
        """
        if local_img == remote_img:
            return True
        return False

    def has_container(self, container):
        con_obj = self.inspect_container(container)
        if con_obj:
            self.input = container
            return con_obj
        return None

    def _inspect_image(self, image):
        try:
            inspect_data = self.d.inspect_image(image)
        except errors.NotFound:
            # We might be looking for something by digest
            if "@sha256:" in image:
                return  self._inspect_image_by_hash(image)
            return None
        return inspect_data

    def _inspect_image_by_hash(self, image):
        input_digest = util.Decompose(image).digest
        all_images = self.get_images(get_all=True)
        for _image in all_images:
            if not _image.repotags:
                continue
            for repo_digest in _image.repotags:
                if no_shaw(input_digest) in repo_digest:
                    return self.d.inspect_image(_image.id)
        return None

    def inspect_image(self, image):
        inspect_data = self._inspect_image(image)
        if inspect_data:
            img_obj = self._make_image(image, inspect_data, deep=True)
            return img_obj
        return None

    def _make_image(self, image, img_struct, deep=False, remote=False):
        img_obj = Image(image, remote=remote)
        img_obj.backend = self

        if not remote:
            img_obj.id = img_struct['Id']
            img_obj.repotags = img_struct['RepoTags']
            img_obj.created = img_struct['Created']
            img_obj.size = img_struct['Size']
            img_obj.virtual_size = img_struct['VirtualSize']
            img_obj.original_structure = img_struct

        if deep:
            img_obj.deep = True
            img_obj.repotags = img_struct['RepoTags']
            img_obj.config = img_struct['Config'] or {}
            img_obj.labels = img_obj.config.get("Labels", None)
            img_obj.os = img_struct['Os']
            img_obj.arch = img_struct['Architecture']
            img_obj.graph_driver = img_struct['GraphDriver']
            img_obj.version = img_obj.get_label('Version')
            img_obj.release = img_obj.get_label('Release')
            img_obj.parent = img_struct['Parent']
            img_obj.original_structure = img_struct
            img_obj.cmd = img_obj.original_structure['Config']['Cmd']
        return img_obj

    def _make_container(self, container, con_struct, deep=False):
        con_obj = Container(container, backend=self)
        con_obj.id = con_struct['Id']
        try:
            con_obj.created = float(con_struct['Created'])
        except ValueError:
            con_obj.created = dateparse(con_struct['Created']).strftime("%F %H:%M") # pylint: disable=no-member
        con_obj.original_structure = con_struct
        try:
            con_obj.name = con_struct['Names'][0]
        except KeyError:
            con_obj.name = con_struct['Name']
        con_obj.input_name = container
        con_obj.backend = self
        try:
            con_obj.command = con_struct['Command']
        except KeyError:
            con_obj.command = con_struct['Config']['Cmd']

        try:
            con_obj.state = con_struct.get('State', None) or con_struct.get['State'].get('Status', None)
        except TypeError:
            # Docker 1.10 on F24 has a different structure
            con_obj.state = con_struct['Status']

        if isinstance(con_obj.state, dict):
            con_obj.state = con_obj.state['Status']
        con_obj.running = True if con_obj.state.lower() in ['true', 'running'] else False

        if deep:
            # Add in the deep inspection stuff
            con_obj.status = con_struct['State']['Status']
            con_obj.image = con_struct['Image']
            con_obj.image_name = con_struct['Config']['Image']
            con_obj.labels = con_struct['Config']['Labels']

        else:
            con_obj.status = con_struct['Status']
            con_obj.image = con_struct['ImageID']
            con_obj.image_name = con_struct['Image']

        return con_obj

    def _inspect_container(self, container):
        try:
            inspect_data = self.d.inspect_container(container)
        except errors.NotFound:
            return None
        return inspect_data

    def inspect_container(self, container):
        inspect_data = self._inspect_container(container)
        if inspect_data:
            return self._make_container(container, inspect_data, deep=True)
        return None

    def get_images(self, get_all=False):
        images = self._get_images(get_all=get_all)
        image_objects = []
        for image in images:
            image_objects.append(self._make_image(image['Id'], image))
        return image_objects

    def make_remote_image(self, image):
        img_obj = self._make_remote_image(image)
        img_obj.populate_remote_inspect_info()
        return img_obj

    def _make_remote_image(self, image):
        return self._make_image(image, None, remote=True)

    def get_containers(self):
        containers = self._get_containers()
        con_objects = []
        for con in containers:
            con_objects.append(self._make_container(con['Id'], con))
        return con_objects

    def _get_images(self, get_all=False, quiet=False, filters=None):
        if filters:
            assert isinstance(filters, dict)
        else:
            filters = {}
        return self.d.images(all=get_all, quiet=quiet, filters=filters)

    def _get_containers(self):
        return self.d.containers(all=True)

    def start_container(self, name):
        if not self.has_container(name):
            raise ValueError("Unable to locate container '{}' in {} backend".format(name, self.backend))
        return self.d.start(name)

    def stop_container(self, con_obj, **kwargs):
        atomic = kwargs.get('atomic')
        args = kwargs.get('args')
        con_obj.stop_args = con_obj.get_label('stop')
        if con_obj.stop_args:
            try:
                cmd = atomic.gen_cmd(con_obj.stop_args.split() + atomic.quote(args.args))
            except TypeError:
                cmd = atomic.gen_cmd(con_obj.stop_args + atomic.quote(args.args))
            cmd = atomic.sub_env_strings(cmd)
            atomic.display(cmd)
            if args.display:
                return 0
            # There should be some error handling around this
            # in case it fails.  And what should then be done?
            return util.check_call(cmd, env=atomic.cmd_env())
        elif args.display:
            return 0

        return self.d.stop(con_obj.id)


    def pull_image(self, image, remote_image_obj, **kwargs):
        assert(isinstance(remote_image_obj, Image))
        debug = kwargs.get('debug', False)
        if image.startswith("dockertar:"):
            path = image.replace("dockertar:", "", 1)
            with open(path, 'rb') as f:
                self.d.load_image(data=f)
            return 0
        fq_name = remote_image_obj.fq_name
        local_image = self.has_image(image)
        if local_image is not None:
            if self.already_has_image(local_image, remote_image_obj):
                raise util.ImageAlreadyExists(image)
        registry, _, _, tag, _ = util.Decompose(fq_name).all
        image = "docker-daemon:{}".format(fq_name)
        if not image.endswith(tag):
            image += ":{}".format(tag)
        if '@sha256:' in image:
            image = image.replace("@sha256:", ":")

        src_creds = kwargs.get('src_creds')
        insecure = True if util.is_insecure_registry(self.d.info()['RegistryConfig'], registry) else False
        trust = Trust()
        trust.discover_sigstore(fq_name)
        util.write_out("Pulling {} ...".format(fq_name))
        util.skopeo_copy("docker://{}".format(fq_name), image, debug=debug, insecure=insecure,
                         policy_filename=trust.policy_filename, src_creds=src_creds)
        return 0

    def delete_container(self, container, force=False):
        return self.d.remove_container(container, force=force)

    def delete_containers_by_image(self, img_obj, force=False):
        containers_by_image = self.get_containers_by_image(img_obj)
        for container in containers_by_image:
            self.delete_container(container.id, force=force)

    def get_containers_by_image(self, img_obj):
        containers = []
        for container in self.get_containers():
            if img_obj.id == container.image:
                containers.append(container)
        return containers


    def _ping(self):
        '''
        Check if the docker daemon is running; if not, exit with
        message and return code 1
        '''
        try:
            self.d.ping()
        except exceptions.ConnectionError:
            raise NoDockerDaemon()

    def delete_image(self, image, force=False):
        assert(image is not None)
        try:
            return self.d.remove_image(image, force=force)
        except errors.APIError as e:
            raise ValueError(str(e))
        except errors.NotFound: # pylint: disable=bad-except-order
            pass
        except HTTPError:
            pass

    def update(self, name, **kwargs):
        debug = kwargs.get('debug', False)
        # A TypeError is thrown if the force keywords is passed in addition to kwargs
        force = kwargs.get('force', False)
        local_only = False
        try:
            remote_image_obj = self.make_remote_image(name)
        except RegistryInspectError:
            # We might be dealing with a local only image
            local_only = True
        if local_only:
            img_obj = kwargs.get('image_object')
            return self.delete_image(img_obj.id, force=True)
        else:
            # pull_image will raise a ValueError if the "latest" image is already present
            self.pull_image(name, remote_image_obj, debug=debug)
        # Only delete containers if a new image is actually pulled.
        img_obj = self.inspect_image(name)
        if force:
            self.delete_containers_by_image(img_obj)

    def prune(self):
        for iid in self.get_dangling_images():
            self.delete_image(iid, force=True)
            util.write_out("Removed dangling Image {}".format(iid))
        return 0

    def get_dangling_images(self):
        return self._get_images(get_all=True, quiet=True, filters={"dangling": True})

    def install(self, image, name, **kwargs):
        pass

    def rpm_install(self, image, name):
        """
        Install system rpm for selected docker image on this system.

        :param image: instance of Atomic.objects.image.Image
        :param name, str, name of the associated container
        :return: instance of ContainerInstallation or None if no rpm was installed
        """
        labels = image.labels or {}
        # we actually don't care about value of the label
        if 'atomic.has_install_files' in labels:
            # we are going to install the system package - the image provides some files to host
            temp_dir = tempfile.mkdtemp()
            try:
                installation = build_rpm_for_docker_backend(image, name, temp_dir, labels)
                RPMHostInstall.install_rpm(installation.destination_path)
            finally:
                shutil.rmtree(temp_dir)
            return installation
        # if the label is not present, we won't install any package on the system

    def uninstall(self, iobject, name=None, **kwargs):
        atomic = kwargs.get('atomic')
        ignore = kwargs.get('ignore')
        assert(isinstance(atomic, Atomic))
        args = atomic.args
        con_obj = None if not name else self.has_container(name)
        # We have a container by that name, need to stop and delete it
        if con_obj:
            if con_obj.running:
                self.stop_container(con_obj)
            self.delete_container(con_obj.id)

        if args.force:
            self.delete_containers_by_image(iobject, force=True)
        else:
            containers_by_image = self.get_containers_by_image(iobject)
            if len(containers_by_image) > 0:
                containers_active = ", ".join([i.name for i in containers_by_image])
                raise ValueError("Containers `%s` are using this image, delete them first or use --force" % containers_active)

        uninstall_command = iobject.get_label('UNINSTALL')
        command_line_args = args.args

        cmd = []
        if uninstall_command:
            try:
                cmd = cmd + uninstall_command
            except TypeError:
                cmd = cmd + uninstall_command.split()
        if command_line_args:
            cmd += command_line_args

        cmd = atomic.gen_cmd(cmd)
        cmd = atomic.sub_env_strings(cmd)
        atomic.display(cmd)
        if args.display:
            return 0


        if cmd:
            result = util.check_call(cmd, env=atomic.cmd_env())
            if result == 0:
                util.InstallData.delete_by_id(iobject.id, ignore=ignore)
            return result

        system_package_nvra = None
        if not ignore:
            install_data = util.InstallData.get_install_data_by_id(iobject.id)
            system_package_nvra = install_data.get("system_package_nvra", None)
        if system_package_nvra:
            RPMHostInstall.uninstall_rpm(system_package_nvra)

        # Delete the entry in the install data
        util.InstallData.delete_by_id(iobject.id, ignore=ignore)
        return self.delete_image(iobject.image, force=args.force)

    def validate_layer(self, layer):
        pass

    def version(self, image):
        return self.get_layers(image)

    def get_layer(self, image):
        _layer = Layer(self.inspect_image(image))
        # Disabling this assignment; not sure where it used and why
        # Enabled it will cause an Attribute Error because image is
        # a str object and has not attrs.  Leaving for historical
        # purposes in case I break something.

        #_layer.remote = image.remote
        return _layer

    def get_layers(self, image):
        layers = []
        layer = self.get_layer(image)
        layers.append(layer)
        while layer.parent:
            layer = self.get_layer(layer.parent)
            layers.append(layer)
        return layers

    def replace_existing_container(self, _iobject, _requested_image, _args):
        if _args.debug:
            util.write_out("Removing the container {} and running with {}".format(_iobject.name,
                                                                                  _requested_image.fq_name))
        self.delete_container(_iobject.id, force=True)
        _iobject = _requested_image
        if _args.command:
            _iobject.user_command = _args.command
        return _iobject

    def run(self, iobject, **kwargs):
        def add_string_or_list_to_list(list_item, value):
            if not isinstance(value, list):
                value = value.split()
            list_item += value
            return list_item

        atomic = kwargs.get('atomic', None)
        args = kwargs.get('args')
        # atomic must be an instance of Atomic
        # args must be a argparse Namespace
        assert(isinstance(atomic, Atomic))
        # The object is a container
        # If container exists and not started, start it
        # If container exists and is started, execute command inside it (docker exec)
        # If container doesn't exist, create one and start it
        if args.command:
            iobject.user_command = args.command
        if isinstance(iobject, Container):
            latest_image = self.inspect_image(iobject.image_name)
            if latest_image.id != iobject.image:
                util.write_out("The '{}' container is using an older version of the installed\n'{}' container image. If "
                               "you wish to use the newer image,\nyou must either create a new container with a "
                               "new name or\nuninstall the '{}' container. \n\n# atomic uninstall --name "
                               "{} {}\n\nand create new container on the {} image.\n\n# atomic update --force "
                               "{}s\n\n removes all containers based on an "
                               "image.".format(iobject.name, iobject.image_name, iobject.name, iobject.name,
                                               iobject.image_name, iobject.image_name, iobject.image_name))

            requested_image = self.has_image(args.image)
            if requested_image is None:
                requested_image = self.has_image(iobject.image)

            if iobject.running:
                if args.replace:
                    iobject = self.replace_existing_container(iobject, requested_image, args)
                    return self.run(iobject, args=args, atomic=atomic)
                return self._running(iobject, args, atomic)
            else:
                # Container with the name exists
                image_id = iobject.image
                if requested_image.id != image_id:
                    if args.replace:
                        iobject = self.replace_existing_container(iobject, requested_image, args)
                    else:
                        try:
                            requested_image_fq_name = requested_image.fq_name
                        except RegistryInspectError:
                            requested_image_fq_name = args.image
                        raise AtomicError("Warning: container '{}' already points to {}\nRun 'atomic run {}' to run "
                                          "the existing container.\nRun 'atomic run --replace '{}' to replace "
                                          "it".format(iobject.name,
                                                      iobject.original_structure['Config']['Image'],
                                                      iobject.name,
                                                      requested_image_fq_name))
                else:
                    if args.replace:
                        iobject = self.replace_existing_container(iobject, requested_image, args)
                    else:
                        return self._start(iobject, args, atomic)

        if iobject.get_label('INSTALL') and not args.ignore and not util.InstallData.image_installed(iobject):
            raise ValueError("The image '{}' appears to have not been installed and has an INSTALL label.  You "
                             "should install this image first.  Re-run with --ignore to bypass this "
                             "error.".format(iobject.name or iobject.image))
        # The object is an image
        command = []
        if iobject.run_command:
            command = add_string_or_list_to_list(command, iobject.run_command)
            if iobject.user_command:
                command = add_string_or_list_to_list(command, iobject.user_command)
            opts_file = iobject.get_label("RUN_OPTS_FILE")
            if opts_file:
                opts_file = atomic.sub_env_strings("".join(opts_file))
                if opts_file.startswith("/"):
                    if os.path.isfile(opts_file):
                        try:
                            atomic.run_opts = open(opts_file, "r").read()
                        except IOError:
                            raise ValueError("Failed to read RUN_OPTS_FILE %s" % opts_file)
                else:
                    raise ValueError("Will not read RUN_OPTS_FILE %s: not absolute path" % opts_file)
        else:
            command += [atomic.docker_binary(), "run"]
            if os.isatty(0):
                command += ["-t"]
            if args.detach:
                command += ["-d"]
            command += atomic.SPC_ARGS if args.spc else atomic.RUN_ARGS
            if iobject.user_command:
                command = add_string_or_list_to_list(command, iobject.user_command)

        if len(command) > 0 and command[0] == "docker":
            command[0] = atomic.docker_binary()

        if iobject.cmd and not iobject.user_command and not iobject.run_command:
            cmd = iobject.cmd if isinstance(iobject.cmd, list) else iobject.cmd.split()
            command += cmd
        command = atomic.gen_cmd(command)
        command = atomic.sub_env_strings(command)
        atomic.display(command)
        if atomic.args.display:
            return

        if not atomic.args.quiet:
            self.check_args(command)
        return util.check_call(command, env=atomic.cmd_env())

    @staticmethod
    def check_args(cmd):
        found_sec_arg = False
        security_args = {
            '--privileged':
                'This container runs without separation and should be '
                'considered the same as root on your system.',
            '--cap-add':
                'Adding capabilities to your container could allow processes '
                'from the container to break out onto your host system.',
            '--security-opt label:disable':
                'Disabling label separation turns off tools like SELinux and '
                'could allow processes from the container to break out onto '
                'your host system.',
            '--security-opt label=disable':
                'Disabling label separation turns off tools like SELinux and '
                'could allow processes from the container to break out onto '
                'your host system.',
            '--net=host':
                'Processes in this container can listen to ports (and '
                'possibly rawip traffic) on the host\'s network.',
            '--pid=host':
                'Processes in this container can see and interact with all '
                'processes on the host and disables SELinux within the '
                'container.',
            '--ipc=host':
                'Processes in this container can see and possibly interact '
                'with all semaphores and shared memory segments on the host '
                'as well as disables SELinux within the container.'
        }

        for sec_arg in security_args:
            if sec_arg in cmd:
                if not found_sec_arg:
                    util.write_out("\nThis container uses privileged "
                                   "security switches:")
                util.write_out("\n\033[1mINFO: {}\033[0m "
                               "\n{}{}".format(sec_arg, " " * 6,
                                               security_args[sec_arg]))
                found_sec_arg = True
        if found_sec_arg:
            util.write_out("\nFor more information on these switches and their "
                           "security implications, consult the manpage for "
                           "'docker run'.\n")

    def _running(self, con_obj, args, atomic):
        requested_image = self.has_image(args.image)
        if requested_image is not None and con_obj.image != requested_image.id:
            requested_image_fq_name = requested_image.fq_name
            raise AtomicError("Warning: container '{}' already points to {}\nRun 'atomic run {}' to run "
                                          "the existing container.\nRun 'atomic run --replace '{}' to replace "
                                          "it".format(con_obj.name,
                                                      con_obj.original_structure['Config']['Image'],
                                                      con_obj.name,
                                                      requested_image_fq_name))
        if con_obj.interactive:
            container_command = con_obj.command if not args.command else args.command
            container_command = container_command if not isinstance(container_command, list) else " ".join(container_command)
            cmd = [atomic.docker_binary(), "exec", "-t", "-i", con_obj.name] + container_command.split()
            if args.display:
                return atomic.display(" ".join(cmd))
            else:
                return util.check_call(cmd, stderr=DEVNULL)
        else:
            command = con_obj.command if not args.command else args.command
            try:
                cmd = [atomic.docker_binary(), "exec", "-t", "-i", con_obj.name] + command
            except TypeError:
                cmd = [atomic.docker_binary(), "exec", "-t", "-i", con_obj.name] + command.split()

            if args.command:
                if args.display:
                    return util.write_out(" ".join(cmd))
                else:
                    return util.check_call(cmd, stderr=DEVNULL)
            else:
                if not args.display:
                    util.write_out("Container is running")

    def _start(self, con_obj, args, atomic):
        exec_error = "Failed to execute the command inside the existing container. In some situations " \
                     "this can happen because the entry point command of the container only runs for " \
                     "a short time. You might want to replace the container by executing your " \
                     "command with --replace. Note any updates to the existing container will be lost"

        if con_obj.interactive:
            if args.command:
                util.check_call([atomic.docker_binary(), "start", con_obj.name], stderr=DEVNULL)
                container_command = args.command if isinstance(args.command, list) else args.command.split()
                try:
                    return util.check_call([atomic.docker_binary(), "exec", "-t", "-i", con_obj.name] + container_command)
                except CalledProcessError as e:
                    if args.debug:
                        util.write_out(str(e))
                    raise AtomicError(exec_error)


            else:
                return util.check_call(
                    [atomic.docker_binary(), "start", "-i", "-a", con_obj.name],
                    stderr=DEVNULL)
        else:
            if args.command:
                util.check_call(
                    [atomic.docker_binary(), "start", con_obj.name],
                    stderr=DEVNULL)
                try:
                    return util.check_call(
                        [atomic.docker_binary(), "exec", "-t", "-i", con_obj.name] +
                        con_obj.command)
                except CalledProcessError as e:
                    if args.debug:
                        util.write_out(str(e))
                    raise AtomicError(exec_error)

            else:
                return util.check_call(
                    [atomic.docker_binary(), "start", con_obj.name],
                    stderr=DEVNULL)

    def tag_image(self, _src, _dest):
        registry, repo, image, tag, _ = util.Decompose(_dest).all
        result = registry
        if repo:
            result += "/{}".format(repo)
        result += "/{}".format(image)
        if result.startswith("/"):
            result = result[1:]
        return self.d.tag(_src, result, tag=tag)




#!/usr/bin/env python
#pylint: skip-file

import dbus
import sys
import subprocess
try:
    from slip.dbus import polkit
except ImportError:
    sys.exit(77)
import dbus.service
import dbus.mainloop.glib
import json

_integration_test_serial = 0

clean_up_tasks = []

# Throw this error to signify at test failure
class AtomicIntegrationTestError(Exception):
    def __init__(self, test_name, error):
        super(AtomicIntegrationTestError, self).__init__("Test '{}' Failed... due to {}".format(test_name, error))


def integration_test(func):
    global _integration_test_serial
    func._integration_test_serial = _integration_test_serial
    _integration_test_serial += 1
    return func


class TestDBus():

    def __init__(self):
        self.bus = dbus.SystemBus()
        self.dbus_object = self.bus.get_object("org.atomic", "/org/atomic/object")
        self.cid = None
        self.registry_cid = None

    @staticmethod
    def run_cmd(cmd):
        if not isinstance(cmd, list):
            cmd = cmd.split()
        return subprocess.check_output(cmd)

    @staticmethod
    def add_cleanup_cmd(cmd):
        assert(isinstance(cmd, str))
        clean_up_tasks.append(cmd.split())

    @staticmethod
    def remove_cleanup_cmd(cmd):
        assert(isinstance(cmd, str))
        clean_up_tasks.remove(cmd.split())

    @staticmethod
    def cleanup():
        for task in clean_up_tasks:
            print("Running clean up: {}".format(" ".join(task)))
            TestDBus.run_cmd(task)

    # Add this decorator to define the method as something that should be
    # tested

    @integration_test
    def test_scan_list(self):
        results = self.dbus_object.ScanList()
        assert(isinstance(results, dbus.String))

    @integration_test
    def test_pull(self):
        TestDBus.add_cleanup_cmd('docker rmi docker.io/library/busybox:1.24')
        assert(self.dbus_object.ImagePull('docker.io/library/busybox:1.24') == 0)

    @integration_test
    def test_pull_bad_image(self):
        try:
            self.dbus_object.ImagePull('docker.io/library/foobar:1234')
            raise ValueError("This test should have resulted in an exception")
        except dbus.DBusException:
            pass

    @integration_test
    def test_update(self):
        TestDBus.add_cleanup_cmd('docker rmi docker.io/busybox:latest')
        self.run_cmd(['docker', 'tag', 'docker.io/library/busybox:1.24', 'docker.io/library/busybox:latest'])
        assert(self.dbus_object.ImageUpdate('docker.io/library/busybox:latest') ==  0)


    @integration_test
    def test_update_already_present(self):
        try:
            self.dbus_object.ImageUpdate('docker.io/library/busybox:latest')
            raise ValueError("This test should have resulted in an exception")
        except dbus.DBusException:
            pass

    @integration_test
    def test_pull_already_present(self):
        try:
            self.dbus_object.ImagePull('docker.io/library/busybox:latest')
            raise ValueError("This should have resulted in an exception")
        except dbus.DBusException:
            pass

    @integration_test
    def test_run(self):
        self.dbus_object.Run('atomic-test-3', 'atomic-dbus-3', False, False, True, '')
        self.cid = TestDBus.run_cmd('docker ps -aq -l').decode('utf-8').rstrip()
        TestDBus.add_cleanup_cmd('docker rm {}'.format(self.cid))
        container_inspect = json.loads(TestDBus.run_cmd('docker inspect {}'.format(self.cid)).decode('utf-8'))[0]
        assert(container_inspect['Name'] == '/atomic-dbus-3')

    @integration_test
    def test_container_delete(self):
        self.dbus_object.ContainersDelete([self.cid])
        TestDBus.remove_cleanup_cmd('docker rm {}'.format(self.cid))

    @integration_test
    def test_container_delete_nonexistent(self):
        try:
            self.dbus_object.ContainersDelete([self.cid])
            raise ValueError("Expected an exception to be raised and was not.")
        except dbus.DBusException:
            pass

    @integration_test
    def test_install(self):
        # Setup
        TestDBus.run_cmd('docker run atomic-test-3')
        t_cid = TestDBus.run_cmd('docker ps -aq -l').decode('utf-8').rstrip()
        TestDBus.run_cmd('docker commit {} dbus-test-3'.format(t_cid))
        TestDBus.run_cmd('docker rm {}'.format(t_cid))
        results = self.dbus_object.Install('dbus-test-3', 'dbus-test-3', False, False, 'docker', False, 'auto', '')
        self.cid = TestDBus.run_cmd('docker ps -aq -l').decode('utf-8').rstrip()
        TestDBus.add_cleanup_cmd('docker rm {}'.format(self.cid))
        TestDBus.add_cleanup_cmd('docker rmi atomic-dbus-3')
        assert(results == 0)

    @integration_test
    def test_uninstall(self):
        results = self.dbus_object.Uninstall('dbus-test-3', '', True, '', True, '')
        TestDBus.remove_cleanup_cmd('docker rm {}'.format(self.cid))
        TestDBus.remove_cleanup_cmd('docker rmi atomic-dbus-3')
        try:
            # The container should have been deleted on uninstall
            TestDBus.run_cmd('docker inspect {}'.format(self.cid))
            raise ValueError("Expected an exception to be raised and was not.")
        except Exception:
            pass
        assert(results == 0)

    @integration_test
    def test_push(self):
        TestDBus.run_cmd('docker pull docker.io/alpine:latest')
        TestDBus.run_cmd('docker tag docker.io/alpine:latest localhost:5000/alpine:latest')
        TestDBus.run_cmd('docker run -d -p 5000:5000 --restart=always --name registry docker.io/library/registry:2')
        self.registry_cid = TestDBus.run_cmd('docker ps -aq -l').decode('utf-8').rstrip()
        TestDBus.add_cleanup_cmd('docker rm -f {}'.format(self.registry_cid))
        TestDBus.add_cleanup_cmd('docker rmi docker.io/library/registry:2')
        TestDBus.add_cleanup_cmd('docker rmi docker.io/alpine:latest')
        TestDBus.add_cleanup_cmd('docker rmi localhost:5000/alpine:latest')
        results = self.dbus_object.ImagePush("localhost:5000/alpine:latest", False, False, False, "", "foo", "bar", "", "", "", "", "", True, False)
        assert(results == 0)

    @integration_test
    def test_push_no_password(self):
        try:
            self.dbus_object.ImagePush("localhost:5000/alpine:latest", False, False, False, "", "foo", "", "", "", "", "", "", True, False)
            raise ValueError("Expected an exception to be raised and was not.")
        except dbus.DBusException:
            pass

    @integration_test
    def test_push_no_username(self):
        try:
            self.dbus_object.ImagePush("localhost:5000/alpine:latest", False, False, False, "", "", "", "", "", "", "", "", True, False)
            raise ValueError("Expected an exception to be raised and was not.")
        except dbus.DBusException:
            pass

    @integration_test
    def test_push_pulp_no_username(self):
        try:
            self.dbus_object.ImagePush("localhost:5000/alpine:latest", True, False, False, "url", "", "", "", "", "", "", "", True, False)
            raise ValueError("Expected an exception to be raised and was not.")
        except dbus.DBusException:
            pass

    @integration_test
    def test_push_pulp_no_url(self):
        try:
            self.dbus_object.ImagePush("localhost:5000/alpine:latest", True, False, False, "", "foo", "bar", "", "", "", "", "", True, False)
            raise ValueError("Expected an exception to be raised and was not.")
        except dbus.DBusException:
            pass

    @integration_test
    def test_stop(self):
        results = self.dbus_object.Stop('{}'.format(self.registry_cid))
        TestDBus.remove_cleanup_cmd('docker rm -f {}'.format(self.registry_cid))
        TestDBus.add_cleanup_cmd('docker rm {}'.format(self.registry_cid))
        assert(results == 0)

    @integration_test
    def test_container_delete(self):
        results = self.dbus_object.ContainersDelete([self.registry_cid])
        assert (results == 0)
        TestDBus.remove_cleanup_cmd('docker rm {}'.format(self.registry_cid))

    @integration_test
    def test_container_delete_nonexistent(self):
        try:
            self.dbus_object.ContainersDelete([self.registry_cid])
            raise ValueError("Expected an exception to be raised and was not.")
        except dbus.DBusException:
            pass

    @integration_test
    def test_image_tag(self):
        try:
            self.dbus_object.ImagePull('docker.io/library/busybox:1.24')
        except:
            pass

        result_tag = self.dbus_object.ImagesTag('docker.io/library/busybox:1.24', 'foobar', 'docker')
        result_delete = self.dbus_object.ImagesDelete(['foobar'], True, False, 'docker')
        assert (result_tag == 0)
        assert (result_delete == 0)

if __name__ == '__main__':

    def get_test_methods(_tb):
        """
        Returns the test methods from above that have the integration_test decorator
        :param _tb: TestDbus instance
        """
        method_names = (n for n in dir(_tb) if not n.startswith('__') and
                        callable(getattr(_tb, n)))
        methods_and_names = ((getattr(_tb, n), n) for n in method_names)

        test_method_names_order = (
            (x[1], getattr(x[0], '_integration_test_serial'))
            for x in methods_and_names
            if '_integration_test_serial' in dir(x[0]))
        test_method_names_ordered = [x[0] for x in
                                     sorted(test_method_names_order,
                                            key=lambda x: x[1])]

        return test_method_names_ordered

    tb = TestDBus()
    test_methods = get_test_methods(tb)
    for test in test_methods:
        _tb = getattr(tb, test)
        try:
            _tb()
        except ValueError as e:
            tb.cleanup()
            raise AtomicIntegrationTestError(test, e)

        print("Test '{}' passed.".format(test))

    tb.cleanup()



from Atomic.util import output_json
import datetime

class Container(object):
    def __init__(self, input_name, backend=None):

        # Required
        self._name = None
        self.id = None
        self._created = None
        self.status = None
        self.input_name = input_name
        self.original_structure = None
        self.deep = False
        self._backend = backend
        self.runtime = backend.backend
        self.image = None
        self.image_name = None
        self._command = None
        self.state = None
        self.vulnerable = False
        self.labels = None
        self._user_command = None
        self.mount_path = None

        # Optional
        self.running = False
        # Instantiate
        self._instantiate()
        self.stop_args = None

    def _instantiate(self):
        self._setup_common()
        return self

    def _setup_common(self):
        # Items common to backends can go here.
        pass

    def get_label(self, label):
        if self.labels:
            return self.labels.get(label.lower(), None) or self.labels.get(label.upper(), None)
        return None

    def dump(self):
        # Helper function to dump out known variables in pretty-print style
        class_vars = dict(vars(self))
        foo = {x: class_vars[x] for x in class_vars if not callable(getattr(self, x)) and not x.startswith('__')
               and not x.endswith('_backend')}
        output_json(foo)

    @property
    def backend(self):
        return self._backend

    @backend.setter
    def backend(self, value):
        self._backend = value

    @property
    def type(self):
        return 'container'

    @property
    def created(self):
        return str(datetime.datetime.fromtimestamp(self._created))

    @property
    def created_raw(self):
        return self._created

    @created.setter
    def created(self, value):
        self._created = value

    @property
    def command(self):
        cmd = self._command if self._command is not None else ['/bin/sh']
        return  cmd

    @command.setter
    def command(self, value):
        self._command = value

    @property
    def interactive(self):
        config = self.original_structure['Config']
        if all([config.get('AttachStdin', False), config.get('AttachStdout', False), config.get('AttachStderr', False)]):
            return True
        return False

    @property
    def name(self):
        return str(self._name)

    @name.setter
    def name(self, value):
        self._name = value[1:] if value[0] == '/' else value

    @property
    def user_command(self):
        return self._user_command

    @user_command.setter
    def user_command(self, value):
        self._user_command = value




class DockerObjectNotFound(ValueError):
    def __init__(self, msg):
        super(DockerObjectNotFound, self).__init__("Unable to associate '{}' with an image or container".format(msg))
        
class NoDockerDaemon(Exception):
    def __init__(self):
        super(NoDockerDaemon, self).__init__("The docker daemon does not appear to be running.")



import re
import os
from . import util
from . import rpmwriter
import tempfile
import shutil

RPM_NAME_PREFIX = "atomic-container"

class RPMHostInstall(object):

    @staticmethod
    def copyfile(src, dest):
        if os.path.isdir(src):
            # add the directory only if it is empty, so we don't create directories that
            # weren't added by us.  Anyway a non empty directory would be created by
            # its children.
            if len(os.listdir(src)) == 0:
                os.mkdir(dest)
        elif os.path.islink(src):
            linkto = os.readlink(src)
            os.symlink(linkto, dest)
        else:
            shutil.copy2(src, dest)

    @staticmethod
    def _do_rename_path(path, rename_files):
        path_split = path.split('/')
        path = ""
        for i in path_split[1:]:
            path = "{}/{}".format(path, i)
            path = rename_files.get(path, path)
        return path

    @staticmethod
    def rm_add_files_to_host(old_installed_files, exports, prefix="/", files_template=None, values=None, rename_files=None):
        # if any file was installed on the host delete it
        if old_installed_files:
            for i in old_installed_files:
                try:
                    os.remove(i)
                except OSError:
                    pass

        if not exports:
            return []

        templates_set = set(files_template or [])

        # if there is a directory hostfs/ under exports, copy these files to the host file system.
        hostfs = os.path.join(exports, "hostfs")
        new_installed_files = []
        if os.path.exists(hostfs):
            for root, dirs, files in os.walk(hostfs):
                rel_root_path = os.path.relpath(root, hostfs)
                if not os.path.exists(os.path.join(prefix, rel_root_path)):
                    os.makedirs(os.path.join(prefix, rel_root_path))
                for f in dirs + files:
                    src_file = os.path.join(root, f)
                    dest_path = os.path.join(prefix, rel_root_path, f)
                    rel_dest_path = os.path.join("/", rel_root_path, f)

                    # If rename_files is set, rename the destination file
                    if rename_files:
                        rel_dest_path = RPMHostInstall._do_rename_path(rel_dest_path, rename_files)
                        dest_path = os.path.join(prefix or "/", os.path.relpath(rel_dest_path, "/"))

                    if os.path.exists(dest_path):
                        util.write_out("File %s already exists." % dest_path, lf="\n")
                        continue

                    if not os.path.exists(os.path.dirname(dest_path)):
                        os.makedirs(os.path.dirname(dest_path))
                    if rel_dest_path in templates_set:
                        with open(src_file, 'r') as src_file_obj:
                            data = src_file_obj.read()
                        util.write_template(src_file, data, values or {}, dest_path)
                        shutil.copystat(src_file, dest_path)
                    else:
                        RPMHostInstall.copyfile(src_file, dest_path)

                    new_installed_files.append(rel_dest_path)
            new_installed_files.sort()  # just for an aesthetic reason in the info file output

        return new_installed_files


    @staticmethod
    def generate_rpm_from_rootfs(rootfs, temp_dir, name, image_id, labels, include_containers_file, display=False, installed_files=None, defaultversion='1'):
        rpm_content = os.path.join(temp_dir, "rpmroot")

        included_rpm = os.path.join(rootfs, "rootfs", "exports", "container.rpm")
        if os.path.exists(included_rpm):
            return included_rpm

        summary = labels.get('atomic.summary', name)
        version = labels.get("atomic.version", defaultversion)
        release = labels.get("atomic.release", image_id)
        license_ = labels.get("atomic.license", "GPLv2")
        url = labels.get("atomic.url")
        requires = labels.get("atomic.requires")
        provides = labels.get("atomic.provides")
        conflicts = labels.get("atomic.conflicts")
        description = labels.get("atomic.description")

        result_dir = os.path.join(temp_dir, "build")
        if not os.path.exists(result_dir):
            os.makedirs(result_dir)

        rpm_name = "atomic-container-%s" % name
        rpm_out = os.path.join(result_dir, "%s.rpm" % rpm_name)
        def split_name_version(pkg):
            r = r"([a-zA-Z0-9_\-\.\+]+)(.*)"
            s = re.search(r, pkg)
            return s.group(1), s.group(2)

        files_to_install = installed_files or []
        if include_containers_file:
            files_to_install.append("/usr/lib/systemd/system/%s.service" % name)
            for root, dirs, files in os.walk(os.path.join(rpm_content, "usr/lib/containers/atomic", name)):
                rel_path = os.path.relpath(root, rpm_content)
                for f in dirs + files:
                    p = os.path.join("/", rel_path, f)
                    files_to_install.append(p)

        with open(rpm_out, "wb") as f, open('/dev/null', 'wb') as devnull:
            writer = rpmwriter.RpmWriter(f, rpm_content, rpm_name, version, release, summary, description or "", license_=license_ or "", url=url or "", stderr=devnull, whitelist=files_to_install)
            if requires is not None:
                for name, version in [split_name_version(i) for i in requires.split(',')]:
                    writer.add_require(name, version)
            if conflicts is not None:
                for name, version in [split_name_version(i) for i in conflicts.split(',')]:
                    writer.add_conflict(name, version)
            if provides is not None:
                for name in provides.split(','):
                    writer.add_provide(name)

            if not display:
                writer.generate()

        return temp_dir

    @staticmethod
    def find_rpm(tmp_dir):
        rpm_file = None
        if tmp_dir == None:
            return None
        for root, _, files in os.walk(os.path.join(tmp_dir, "build")):
            if rpm_file:
                break
            for f in files:
                if f.endswith('.rpm'):
                    rpm_file = os.path.join(root, f)
                    break
        return rpm_file

    @staticmethod
    def generate_rpm(name, image_id, labels, exports, destination, values=None, installed_files=None, installed_files_template=None, rename_files=None, defaultversion='1'):

        if values == None:
            values = {}

        # If rpm.spec or rpm.spec.template exist, copy them to the checkout directory, processing the .template version.
        if os.path.exists(os.path.join(exports, "rpm.spec.template")):
            with open(os.path.join(exports, "rpm.spec.template"), "r") as f:
                spec_content = f.read()
            util.write_template("rpm.spec.template", spec_content, values, os.path.join(destination, "rpm.spec"))
        elif os.path.exists(os.path.join(exports, "rpm.spec")):
            shutil.copyfile(os.path.join(exports, "rpm.spec"), os.path.join(destination, "rpm.spec"))

        temp_dir = tempfile.mkdtemp()
        orig_name = dest_path = None
        try:
            rpm_content = os.path.join(temp_dir, "rpmroot")
            rootfs = os.path.join(rpm_content, "usr/lib/containers/atomic", name)
            os.makedirs(rootfs)
            if installed_files is None:
                installed_files = RPMHostInstall.rm_add_files_to_host(None, exports, rpm_content, files_template=installed_files_template, values=values, rename_files=rename_files)
            rpm_root = RPMHostInstall.generate_rpm_from_rootfs(destination, temp_dir, name, image_id, labels, include_containers_file=False, installed_files=installed_files, defaultversion=defaultversion)
            rpm_file = RPMHostInstall.find_rpm(rpm_root)
            if rpm_file:
                orig_name = "atomic-container-{}.rpm".format(name)
                dest_path = os.path.join(destination, orig_name)
                if os.path.exists(dest_path):
                    os.unlink(dest_path)
                shutil.move(rpm_file, dest_path)
        finally:
            shutil.rmtree(temp_dir)
        return orig_name, dest_path, installed_files

    @staticmethod
    def install_rpm(rpm):
        """
        :param rpm_file: str, name of the rpm to install, is passed to dnf/yum
        :return: None
        """
        if os.path.exists("/run/ostree-booted"):
            raise ValueError("This doesn't work on Atomic Host yet")
        elif os.path.exists("/usr/bin/dnf"):
            util.check_call(["dnf", "install", "-y", rpm])
        else:
            util.check_call(["yum", "install", "-y", rpm])

    @staticmethod
    def uninstall_rpm(rpm):
        """
        :param rpm: str, name of the rpm to uninstall, is passed to dnf/yum
        :return: None
        """
        if os.path.exists("/run/ostree-booted"):
            raise ValueError("This doesn't work on Atomic Host yet")
        elif os.path.exists("/usr/bin/dnf"):
            util.check_call(["dnf", "remove", "-y", rpm])
        else:
            util.check_call(["yum", "remove", "-y", rpm])



import os
import json
try:
    import ConfigParser as configparser
except ImportError:  # py3 compat
    import configparser # pylint: disable=import-error

import requests

from . import util


util.urllib3_disable_warnings()


def push_image_to_satellite(image, server_url, username, password,
                            verify_ssl, docker_client, activation_key,
                            repo_id, debug=False):
    if not image:
        raise ValueError("Image required")
    parts = image.split("/")
    if parts > 1:
        if parts[0].find(".") != -1:
            server_url = parts[0]
            image = ("/").join(parts[1:])
    if not server_url:
        raise ValueError("Satellite server url required")

    if not server_url.startswith("http"):
        server_url = "https://" + server_url

    try:
        sat = SatelliteServer(server_url=server_url, username=username,
                              password=password, verify_ssl=verify_ssl,
                              docker_client=docker_client, debug=debug)
    except IOError as e:
        raise IOError('Failed to initialize Satellite: {0}'.format(e))
    if not sat.is_repo(repo_id):
        raise IOError("""Invalid Repository ID: {0}.  Please create that repository
and try again, or input a different ID.""".format(repo_id).replace('\n', ' '))
    keyData = sat.get_data(repo_id, activation_key)
    content_view_id = keyData.get("content_view_id")
    try:
        util.write_out('Uploading image "{0}" to server "{1}"'.format(
                      image, server_url))
        sat.upload_docker_image(image, repo_id)
        util.write_out("")
    except IOError as e:
        raise IOError('Failed to upload image: {0}'.format(e))
    sat.publish_view(content_view_id, repo_id)
    print("Push Complete")


class SatelliteServer(object):
    # Interact with Satellite API
    def __init__(self, server_url, username, password, verify_ssl,
                 docker_client, debug=False):
        self._server_url = server_url
        self._username = username
        self._password = password
        self._verify_ssl = verify_ssl
        self._docker_client = docker_client
        self._web_distributor = "docker_web_distributor_name_cli"
        self._export_distributor = "docker_export_distributor_name_cli"
        self._importer = "docker_importer"
        self._export_dir = "/var/www/pub/docker/web/"
        self._unit_type_id = "docker_image"
        self._chunk_size = 1048576  # 1 MB per upload call
        self._debug = debug

    def _call_satellite(self, url, req_type='get', payload=None):
        """This function handles requests to the Satellite Server"""
        proxies = util.get_proxy()
        if req_type == 'get':
            if (self._debug):
                print('Calling Satellite URL "{0}"'.format(url))
            r = requests.get(url, auth=(self._username, self._password),
                             verify=self._verify_ssl, proxies=proxies)
        elif req_type == 'post':
            if (self._debug):
                print('Posting to Satellite URL "{0}"'.format(url))
                if payload:
                    print('Satellite HTTP payload:\n{0}'.format(
                          json.dumps(payload, indent=2)))
            r = requests.post(url, auth=(self._username, self._password),
                              data=json.dumps(payload),
                              verify=self._verify_ssl,
                              proxies=proxies)
        elif req_type == 'post-nodata':
            if (self._debug):
                print('Posting to Satellite URL "{0}". No data sent.'.format(
                    url))
            header = {'Content-Type': 'application/json'}
            r = requests.post(url, auth=(self._username, self._password),
                              headers=header, data=json.dumps(payload),
                              verify=self._verify_ssl,
                              proxies=proxies)
        elif req_type == 'put':
            if self._debug:
                print('Putting to Satellite URL "{0}"'.format(url))
            r = requests.put(url, auth=(self._username, self._password),
                             data=payload, verify=self._verify_ssl,
                             proxies=proxies)
        elif req_type == 'put-jsonHead':
            if self._debug:
                print('Putting with json header to Satellite URL "{0}"'
                      .format(url))
            header = {'Content-Type': 'application/json'}
            r = requests.put(url, auth=(self._username, self._password),
                             headers=header, data=json.dumps(payload),
                             verify=self._verify_ssl, proxies=proxies)
        elif req_type == 'put-multi-part':
            if self._debug:
                print('Multi-Part Putting to Satellite URL "{0}"'.format(url))
            header = {
                'multipart': True,
                'accept': 'application/json;version=2',
                'content-type': 'multipart/form-data'
            }
            r = requests.put(url, auth=(self._username, self._password),
                             headers=header, data=payload,
                             verify=self._verify_ssl,
                             proxies=proxies)
        elif req_type == 'delete':
            if self._debug:
                print('Delete call to Satellite URL "{0}"'.format(url))
            header = {'Content-Type': 'application/json'}
            r = requests.delete(url, auth=(self._username, self._password),
                                headers=header, verify=self._verify_ssl,
                                proxies=proxies)
        else:
            raise IOError('Invalid value of "req_type" parameter: {0}'
                             .format(req_type))
        if self._debug:
            print(r)
        try:
            r_json = r.json()
        except IOError:
            # some requests don't return a json object
            return None

        if ('errors' in r_json):
            util.write_err('Error message from Satellite response:\n{0}'
                             .format(r_json['errors']))
        if 'spawned_tasks' in r_json:
            for task in r_json['spawned_tasks']:
                if self._debug:
                    print('Checking status of spawned task {0}'.format(
                        task['task_id']))
                self._call_satellite('{0}/{1}'.format(self._server_url,
                                                      task['_href']))
        return r_json

    # It looks like, while we can't actually get the activation key
    # from the text on the satellite page, it is in the URL of the
    # page. Same for repo id with repo page.  As in,
    # https://sat6-atomic.refarch.bos.redhat.com/activation_keys/
    # {The activation key number is here}/info
    def get_data(self, repo_id, activation_key):
        url = '{0}/katello/api/repositories/{1}'.format(
            self._server_url, repo_id)
        r_json = self._call_satellite(url)
        keyData = {}
        keyData['org_id'] = r_json.get("organization").get("id")
        keyData['product_id'] = r_json.get("product").get("id")
        url = '{0}/katello/api/activation_keys/{1}'.format(
            self._server_url, activation_key)
        r2_json = self._call_satellite(url)
        keyData['content_view_id'] = r2_json.get("content_view_id")
        if self._debug:
            print("key data is {0}".format(keyData))
        return keyData

    @property
    def status(self):
        """Return Satellite server status"""
        if self._debug:
            print('Verifying Satellite server status')
        return self._call_satellite('{0}/api/v2/status/'.format(
            self._server_url))

    def is_repo(self, repo_id):
        """Return true if repo exists"""
        url = '{0}/katello/api/repositories/{1}'.format(
            self._server_url, repo_id)
        if self._debug:
            print('Verifying satellite repository "{0}"'.format(repo_id))
        r_json = self._call_satellite(url)
        if int(repo_id) == r_json.get('id'):
            if self._debug:
                print("Yes it was a repo")
            return True
        else:
            return False

    def _upload_id(self, repo_id):
        """Get a satellite upload ID"""
        url = '{0}/katello/api/repositories/{1}/content_uploads'.format(
            self._server_url, repo_id)
        r_json = self._call_satellite(url, "post-nodata")
        if 'error' in r_json:
            raise IOError('Unable to get a satellite upload ID')
        return r_json.get('upload_id')

    def upload_docker_image(self, image, repo_id):
        """Upload image to  repository"""
        if self._debug:
            print("Getting an upload id")
        upload_id = self._upload_id(repo_id)
        if self._debug:
            print('Uploading image using ID "{0}"'.format(upload_id))
            print('\nUploading image "{0}"'.format(image))
        self._upload_docker_image(image, repo_id, upload_id)
        self._import_upload(upload_id, repo_id)
        self._delete_upload_id(upload_id, repo_id)

    def _upload_docker_image(self, image, repo_id, upload_id):
        if self._debug:
            print("Beginning to upload the image")
        offset = 0
        image_stream = self._docker_client.get_image(image)
        while True:
            content = image_stream.read(self._chunk_size)
            if not content:
                break
            url = "{0}/katello/api/repositories/{1}/content_uploads/{2}".format(self._server_url, repo_id, upload_id)
            util.write_out(".", "")
            payload = {
                'offset': offset,
                'content': content
            }
            r_json = self._call_satellite(url, "put-multi-part", payload)
            if (r_json is not None):
                if ('errors' in r_json):
                    raise IOError("Unable to upload image.  Error:{0}"
                                    .format(r_json.get("errors")))
            offset += self._chunk_size
        image_stream.close()
        if self._debug:
            print("Finished uploading the image data")

    def _delete_upload_id(self, upload_id, repo_id):
        """Delete upload request ID"""
        delete_url = "{0}/katello/api/repositories/{1}/content_uploads/{2}".format(self._server_url, repo_id, upload_id)
        self._call_satellite(delete_url, "delete")
        if self._debug:
            print("Successful Deletion")

    def _import_upload(self, upload_id, repo_id):
        """Import uploaded content"""
        url = '{0}/katello/api/repositories/{1}/import_uploads'.format(
            self._server_url, repo_id)
        if self._debug:
            print('Importing satellite upload {0} into {1}'.format(
                upload_id, repo_id))
        payload = {
            "upload_ids": [upload_id]
            # may need to make the id into a string.  Unclear.
        }
        r_json = self._call_satellite(url, "put-jsonHead", payload)
        if (r_json is not None):
            if ('errors' in r_json):
                raise IOError('Unable to import satellite content into {0}'
                                .format(repo_id))

    def publish_view(self, content_id, repo_id):
        """Publish satellite repository to satellite web server"""
        url = '{0}/katello/api/content_views/{1}/publish/'.format(
            self._server_url, content_id)
        r_json = self._call_satellite(url, "post-nodata")
        if (r_json is not None):
            if ('errors' in r_json):
                raise IOError('Unable to publish satellite repo "{0}"'
                                .format(repo_id))


class SatelliteConfig(object):
    def __init__(self):
        self.c = configparser.ConfigParser()
        self.config_file = os.path.expanduser("~/.satellite/admin.conf")
        self.c.read(self.config_file)
        self.url = self._get("server", "url")
        self.username = self._get("auth", "username")
        self.password = self._get("auth", "password")
        self.verify_ssl = self._getboolean("server", "verify_ssl")
    # Satellite configuration file [optional]:
    # 1. look in (or create) ~/.satellite/admin.conf
    # configuration contents:
    # [server]
    # host = <satellite-server-hostname.example.com>
    # verify_ssl = false
    #
    # # optional auth section
    # [auth]
    # username: <user>
    # password: <pass>
    def _get(self, section, val):
        try:
            return self.c.get(section, val)
        except (configparser.NoSectionError, configparser.NoOptionError):
            return None
        except IOError as e:
            raise IOError("Satellite Bad Value for {0} in {1}. {2}".format(
                val, self.config_file, e))

    def _getboolean(self, section, val):
        try:
            return self.c.getboolean(section, val)
        except (configparser.NoSectionError, configparser.NoOptionError):
            return True
        except IOError as e:
            raise IOError("Satellite Bad Value for {0} in {1}. {2}".format(
                val, self.config_file, e))

    def config(self):
        return {"url": self.url, "verify_ssl": self.verify_ssl, "username": self.username, "password": self.password}



from . import Atomic
from . import util
import sys
from Atomic.backendutils import BackendUtils

ATOMIC_CONFIG = util.get_atomic_config()
storage = ATOMIC_CONFIG.get('default_storage', "docker")

class Delete(Atomic):
    def __init__(self):
        super(Delete, self).__init__()
        self.be = None

    def delete_image(self):
        """
        Mark given image(s) for deletion from registry
        :return: 0 if all images marked for deletion, otherwise 2 on any failure
        """
        if self.args.debug:
            util.write_out(str(self.args))

        if (len(self.args.delete_targets) > 0 and self.args.all) or (len(self.args.delete_targets) < 1 and not self.args.all):
            raise ValueError("You must select --all or provide a list of images to delete.")

        beu = BackendUtils()
        delete_objects = []

        # We need to decide on new returns for dbus because we now check image
        # validity prior to executing the delete.  If there is going to be a
        # failure, it will be here.
        #
        # The failure here is basically that it couldnt verify/find the image.

        if self.args.all:
            if self.args.storage:
                self.be = beu.get_backend_from_string(self.args.storage)
                delete_objects = self.be.get_images(get_all=True)
            else:
                delete_objects = beu.get_images(get_all=True)
        else:
            for image in self.args.delete_targets:
                _, img_obj = beu.get_backend_and_image_obj(image, str_preferred_backend=self.args.storage or storage, required=True if self.args.storage else False)
                delete_objects.append(img_obj)

        if self.args.remote:
            return self._delete_remote(self.args.delete_targets)
        if len(delete_objects) == 0:
            raise ValueError("No images to delete.")
        _image_names = []
        for del_obj in delete_objects:
            if del_obj.repotags:
                _image_names.append(len(del_obj.repotags[0]))
            else:
                _image_names.append(len(del_obj.id))
        max_img_name = max(_image_names) + 2

        if not self.args.assumeyes:
            util.write_out("Do you wish to delete the following images?\n")
        else:
            util.write_out("The following images will be deleted.\n")

        two_col = "   {0:" + str(max_img_name) + "} {1}"
        util.write_out(two_col.format("IMAGE", "STORAGE"))
        for del_obj in delete_objects:
            image = None if not del_obj.repotags else del_obj.repotags[0]
            if image is None or "<none>" in image:
                image = del_obj.id[0:12]
            util.write_out(two_col.format(image, del_obj.backend.backend))
        if not self.args.assumeyes:
            confirm = util.input("\nConfirm (y/N) ")
            confirm = confirm.strip().lower()
            if not confirm in ['y', 'yes']:
                util.write_err("User aborted delete operation for {}".format(self.args.delete_targets or "all images"))
                sys.exit(2)

        # Perform the delete
        for del_obj in delete_objects:
            del_obj.backend.delete_image(del_obj.input_name, force=self.args.force)

        # We need to return something here for dbus
        return 0

    def prune_images(self):
        """
        Remove dangling images from registry
        :return: 0 if all images deleted or no dangling images found
        """

        if self.args.debug:
            util.write_out(str(self.args))

        beu = BackendUtils()
        for backend in beu.available_backends:
            be = backend()
            be.prune()

        return 0

    def _delete_remote(self, targets):
        results = 0
        for target in targets:
            # _convert_to_skopeo requires registry v1 support while delete requires v2 support
            # args, img = self.syscontainers._convert_to_skopeo(target)

            args = []
            if "http:" in target:
                args.append("--insecure")

            for i in ["oci:", "http:", "https:"]:
                img = target.replace(i, "docker:")

            if not img.startswith("docker:"):
                img = "docker://" + img

            try:
                util.skopeo_delete(img, args)
                util.write_out("Image {} marked for deletion".format(img))
            except ValueError as e:
                util.write_err("Failed to mark Image {} for deletion: {}".format(img, e))
                results = 2
        return results





from . import Atomic
import argparse
import tempfile
from pydoc import pager
import os
from . import mount
import sys
from . import util

def cli(subparser, hidden=False):
    #atomic help
    if hidden:
        helpp = subparser.add_parser("help", argument_default=argparse.SUPPRESS)

    else:
        helpp = subparser.add_parser("help",
                                     help=_("display help associated with the image"),
                                     epilog="atomic help 'image'")

    helpp.set_defaults(_class=AtomicHelp, func='help_tty')
    helpp.add_argument("image", help=_("Image ID or name"))


GROFF_BINARY = "/usr/bin/groff"


class AtomicHelp(Atomic):
    ATOMIC_DIR="/run/atomic"

    def __init__(self):
        super(AtomicHelp, self).__init__()
        if not os.path.exists(self.ATOMIC_DIR):
            os.makedirs(self.ATOMIC_DIR)

        # possible filenames for help file inside container
        self.help_file_candidates = (
            # help filename inside image, command to process the help file before displaying
            ("help.1", (GROFF_BINARY, '-t', '-man', '-Tascii')),
            ("README.md", None)
        )

        self.docker_object = None
        self.is_container = True
        self.use_pager = True
        self.image = None
        self.inspect = None
        self.enc = sys.getdefaultencoding()

    def help_tty(self):
        result = self.help()
        if not sys.stdout.isatty():
            util.write_out("\n{}\n".format(result))
        else:
            # Call the pager
            os.environ['PAGER'] = '/usr/bin/less -R'
            pager(result)

    def help(self):
        """
        Displays help text for a container.
        :return: None
        """
        self.docker_object = self.args.image
        docker_id = self.get_input_id(self.docker_object)
        self.inspect = self._inspect_container(docker_id)
        if self.inspect is None:  # docker_id is an image
            self.inspect = self._inspect_image(docker_id)
            self.is_container = False
        else:
            # The docker object is a container, need to set
            # its image
            self.image = self.inspect['Image']

        try:
            return self.man_help(docker_id)
        except ValueError:
            # Check if "help" label is provided
            # The label contains command which will be executed
            help_cmd = self._get_args('HELP')
            if help_cmd:
                return self.alt_help(help_cmd)
            else:
                raise ValueError("There is no help for {}.".format(self.docker_object))

    def man_help(self, docker_id):
        """
        Display the help for a container or image using the default
        method of displaying a man formatted page
        :param docker_id: docker object to get help for
        :return: None
        """
        mount_location = tempfile.mkdtemp(prefix=self.ATOMIC_DIR)
        try:
            dm = mount.DockerMount(mount_location)
            with mount.MountContextManager(dm, docker_id):
                # defined b/c of pylint, this is not needed due to for & else construct
                candidate_file = None
                candidate_preprocessor = None
                for candidate_file, candidate_preprocessor in self.help_file_candidates:
                    # overlay
                    help_path = os.path.join(dm.mountpoint, candidate_file)
                    if not os.path.exists(help_path):
                        # devicemapper
                        help_path = os.path.join(dm.mountpoint, 'rootfs', candidate_file)
                    if os.path.exists(help_path):
                        break
                else:
                    # not found
                    raise ValueError(
                        "Unable to find help file for {}.\nTried these files {}.".format(
                            self.docker_object, [x[0] for x in self.help_file_candidates])
                    )

                with open(help_path, "r") as help_file:
                    if candidate_preprocessor:
                        if not os.path.exists(candidate_preprocessor[0]):
                            raise IOError(
                                "Cannot display help file {} for {}: {} unavailable".format(
                                    candidate_file, self.docker_object, candidate_preprocessor)
                            )

                        return util.check_output(candidate_preprocessor, stdin=help_file)
                    return help_file.read()
        finally:
            os.rmdir(mount_location)

    def alt_help(self, help_cmd):
        """
        Returns help when the HELP LABEL override is being used.
        :return: None
        """
        cmd = self.gen_cmd(help_cmd)
        cmd = self.sub_env_strings(cmd)
        self.display(cmd)
        return util.check_output(cmd, env=self.cmd_env()).decode(self.enc)



"""
export docker images, containers and volumes into a filesystem directory.
"""
import os

from .client import AtomicDocker
from . import util

ATOMIC_LIBEXEC = os.environ.get('ATOMIC_LIBEXEC', '/usr/libexec/atomic')

def export_docker(graph, export_location, force):
    """
    This is a wrapper function for exporting docker images, containers
    and volumes.
    """

    if not os.path.isdir(export_location):
        os.makedirs(export_location)

    with AtomicDocker() as client:
        dangling_images = client.images(filters={"dangling":True}, quiet=True)
        if any(dangling_images):
            if not force:
                choice = util.input("There are dangling images in your system. Would you like atomic to prune them [y/N]")
                choice = choice.strip().lower()
                if not choice in ['y', 'yes']:
                    raise ValueError("Please delete dangling images before running atomic storage export")

            util.write_out("Deleting dangling images")
            util.check_call([util.default_docker(), "rmi", "-f"]+dangling_images)

        #Save the docker storage driver
        storage_driver = client.info()["Driver"]
        filed = open(export_location+"/info.txt", "w")
        filed.write(storage_driver)
        filed.close()

        #export docker images
        export_images(export_location)
        #export docker containers
        export_containers(graph, export_location)
        #export docker volumes
        export_volumes(graph, export_location)

        util.write_out("atomic export completed successfully")

def export_images(export_location):
    """
    Method for exporting docker images into a filesystem directory.
    """
    if not os.path.isdir(export_location + "/images"):
        os.makedirs(export_location + "/images")

    images = {}
    with AtomicDocker() as client:
        for image in client.images():
            Id, tags = image["Id"], image["RepoTags"]

            if '<none>:<none>' in tags:
                continue
            if Id not in images:
                images[Id] = []
            images[Id].extend(tags)

    for Id in images:
        tags = " ".join(images[Id])
        util.write_out("Exporting image: {0}".format(Id[:12]))
        with open(export_location + '/images/' + Id, 'w') as f:
            util.check_call([util.default_docker(), "save", tags], stdout=f)

def export_containers(graph, export_location):
    """
    Method for exporting docker containers into a filesystem directory.
    """
    if not os.path.isdir(export_location + "/containers"):
        os.makedirs(export_location + "/containers")

    with AtomicDocker() as client:
        for container in client.containers(all=True):
            Id = container["Id"]

            util.write_out("Exporting container: {0}".format(Id[:12]))
            util.check_call([ATOMIC_LIBEXEC + '/migrate.sh',
                             'export',
                             '--container-id=' + Id[:12],
                             '--graph=' + graph,
                             '--export-location=' + export_location])

def tar_create(srcdir, destfile):
    util.check_call(['/usr/bin/tar', '--create', '--gzip', '--selinux',
                           '--file', destfile, '--directory', srcdir, '.'])


def export_volumes(graph, export_location):
    """
    Method for exporting docker volumes into a filesystem directory.
    """
    if not os.path.isdir(export_location + "/volumes"):
        os.makedirs(export_location + "/volumes")

    util.write_out("Exporting volumes")
    tar_create(srcdir=graph + '/volumes',
               destfile=export_location + '/volumes/volumeData.tar.gz')

    if os.path.isdir(graph + "/vfs"):
        tar_create(srcdir=graph + '/vfs',
                   destfile=export_location + '/volumes/vfsData.tar.gz')



#pylint: skip-file
import os
import unittest

import yaml


class TestAtomicUnit(unittest.TestCase):
    """
    Checks that the configuration file provided is valid.
    """

    def setUp(self):
        """
        Provide the conf file contents freshly for each test.
        """
        self.conf = open(os.path.sep.join(['atomic.conf']), 'r')

    def tearDown(self):
        """
        Close the config file after every test.
        """
        self.conf.close()

    def test_config_file_is_valid_yaml(self):
        """
        Verifies atomic.conf is valid YAML.
        """
        self.assertEquals(type(yaml.safe_load(self.conf)), dict)

    def test_config_file_is_valid_yaml_with_items_uncommented(self):
        """
        Verifies atomic.conf is valid YAML when examples are uncommented.
        """
        # If a command line has a space after it, it's a comment
        # If a comment line has no space after it, it's an example
        uncommented = []
        for line in self.conf.readlines():
            if line.startswith('#') and len(line) > 2 and line[1] != ' ':
                line = line[1:]
            uncommented.append(line)
        self.assertEquals(type(yaml.safe_load('\n'.join(uncommented))), dict)


if __name__ == '__main__':
    unittest.main()



import getpass

from . import util
from . import pulp
from . import satellite

try:
    from . import Atomic
except ImportError:
    from atomic import Atomic # pylint: disable=relative-import

ATOMIC_CONFIG = util.get_atomic_config()
REGISTRY_TYPE_CHOICES = ['docker', 'atomic']

def cli(subparser):
    signer = ATOMIC_CONFIG.get('default_signer', None)
    # atomic push
    pushp = subparser.add_parser(
        "push", aliases=['upload'], help=_("push latest image to repository"),
        epilog="push the latest specified image to a repository.")
    pushp.set_defaults(_class=Push, func='push')
    pushp.add_argument("--anonymous",
                     default=False,
                     action="store_true",
                     help=_("Push images without a username or password"))
    # making it so we cannot call both the --pulp and --satellite commands
    # at the same time (mutually exclusive)
    pushgroup = pushp.add_mutually_exclusive_group()
    pushgroup.add_argument("--pulp",
                           default=False,
                           action="store_true",
                           help=_("push image using pulp"))
    pushgroup.add_argument("--satellite",
                           default=False,
                           action="store_true",
                           help=_("push image using Satellite"))

    pushp.add_argument("--verify_ssl",
                         default=None,
                         action="store_true",
                         help=_("flag to verify ssl of registry"))
    pushp.add_argument("-U", "--url",
                         dest="url",
                         default=None,
                         help=_("URL for remote registry"))
    pushp.add_argument("-u", "--username",
                         default=None,
                         dest="username",
                         help=_("Username for remote registry"))
    pushp.add_argument("-p", "--password",
                         default=None,
                         dest="password",
                         help=_("Password for remote registry"))
    pushp.add_argument("image", help=_("container image"))
    pushp.add_argument("-a", "--activation_key",
                         default=None,
                         dest="activation_key",
                         help=_("Activation Key"))
    pushp.add_argument("-r", "--repository_id",
                         default=None,
                         dest="repo_id",
                         help=_("Repository ID"))
    pushp.add_argument("-t", "--type", dest="reg_type", default="docker",
                       choices=REGISTRY_TYPE_CHOICES,
                       help=_("Registry type"))
    pushp.add_argument("--sign-by", dest="sign_by", default=signer,
                       help=_("Name of the signing key. Currently %s, "
                              "default can be defined in /etc/atomic.conf" % signer))
    pushp.add_argument("-g", "--gnupghome",
                       default=None,
                       dest="gnupghome",
                       help=_("Set the GNUPGHOME environment variable to "
                              "use an alternate user's GPG keyring for signing. "
                              "Useful when running with sudo, "
                              "e.g. set to '~/.gnupg'."))
    pushp.add_argument("--insecure", dest="insecure", default=False,
                       action='store_true', help=_("Do not check registry certificates"))
    # pushp.add_argument("--activation_key_name",
    #                      default=None,
    #                      dest="activation_key_name",
    #                      help=_("Activation Key Name"))
    # pushp.add_argument("--repo_name", "--repository_name",
    #                      default=None,
    #                      dest="repo_name",
    #                      help=_("Repository Name"))
    # pushp.add_argument("--org_name", "--organization_name",
    #                      default=None,
    #                      dest="org_name",
    #                      help=_("Organization Name"))

class Push(Atomic):
    def __init__(self, policy_filename=None):
        """
        :param policy_filename: override policy filename
        """
        super(Push, self).__init__()
        self.policy_filename=policy_filename

    def push(self):
        def prompt():
            if not self.args.username:
                self.args.username = util.input("Registry username: ")

            if not self.args.password:
                self.args.password = getpass.getpass("Registry password: ")

        self.ping()
        if self.args.debug:
            util.write_out(str(self.args))

        # This allows a user to turn off signing when a global
        # sign_by has been defined in /etc/atomic.conf and saves
        # us from having to define something like --no-sign
        if self.args.sign_by == "None":
            self.args.sign_by = None

        # Priority order:
        # If user passes in a password/username/url/ssl flag, use that
        # If not, read from the config file
        # If still nothing, ask again for registry user/pass
        if self.args.pulp:
            config = pulp.PulpConfig().config()

        if self.args.satellite:
            config = satellite.SatelliteConfig().config()

        if (self.args.satellite | self.args.pulp):
            if not self.args.username:
                self.args.username = config["username"]
            if not self.args.password:
                self.args.password = config["password"]
            if not self.args.url:
                self.args.url = config["url"]
            if self.args.verify_ssl is None:
                self.args.verify_ssl = config["verify_ssl"]

        if self.args.verify_ssl is None:
            self.args.verify_ssl = False

        if (self.args.satellite | self.args.pulp):
            prompt()
            if not self.args.url:
                self.args.url = util.input("URL: ")

        sign = True if self.args.sign_by else False

        if self.args.pulp:
            return pulp.push_image_to_pulp(self.image, self.args.url,
                                           self.args.username,
                                           self.args.password,
                                           self.args.verify_ssl,
                                           self.d)

        if self.args.satellite:
            if not self.args.activation_key:
                self.args.activation_key = util.input("Activation Key: ")
            if not self.args.repo_id:
                self.args.repo_id = util.input("Repository ID: ")
            return satellite.push_image_to_satellite(self.image,
                                                     self.args.url,
                                                     self.args.username,
                                                     self.args.password,
                                                     self.args.verify_ssl,
                                                     self.d,
                                                     self.args.activation_key,
                                                     self.args.repo_id,
                                                     self.args.debug)

        else:
            reg, _, _, tag, _ = util.Decompose(self.image).all
            # Check if any local tokens exist
            if reg not in [x for x in self.get_local_tokens()] and not self.args.anonymous:
                # If we find a token for the registry, we don't
                # prompt for a username or password
                prompt()

            if not tag:
                raise ValueError("The image being pushed must have a tag")

            local_image = "docker-daemon:{}".format(self.image)
            if self.args.reg_type == "atomic":
                remote_image = "atomic:{}".format(self.image)
            else:
                remote_image = "docker://{}".format(self.image)

            if sign and self.args.debug:
                util.write_out("\nSigning with '{}'\n".format(self.args.sign_by))

            if self.args.insecure:
                insecure = True
            else:
                insecure = True if util.is_insecure_registry(self.d.info()['RegistryConfig'], reg) else False
            # We must push the file to the registry first prior to performing a
            # local signature because the manifest file must be on the registry
            return_code = util.skopeo_copy(local_image, remote_image, debug=self.args.debug,
                                           policy_filename=self.policy_filename,
                                           sign_by=self.args.sign_by if sign else None, insecure=insecure,
                                           username=self.args.username,
                                           password=self.args.password,
                                           gpghome=self.args.gnupghome)

            if return_code != 0:
                raise ValueError("Pushing {} failed.".format(self.image))
            if self.args.debug:
                util.write_out("Pushed: {}".format(self.image))
            return 0



"""
import docker images, containers and volumes from a filesystem directory.
"""
import os
import sys

from . import util


ATOMIC_LIBEXEC = os.environ.get('ATOMIC_LIBEXEC', '/usr/libexec/atomic')

def import_docker(graph, import_location, assumeyes):
    """
    This is a wrapper function for importing docker images, containers
    and volumes.
    """

    if not os.path.isdir(import_location):
        sys.exit("{0} does not exist".format(import_location))
    #import docker images
    import_images(import_location)
    #import docker containers
    import_containers(graph, import_location)
    #import docker volumes
    import_volumes(graph, import_location)

    util.write_out("atomic import completed successfully")
    confirm = ""
    if assumeyes:
        confirm = "yes"
    else:
        confirm = util.input("Would you like to cleanup (rm -rf {0}) the temporary directory [y/N]"
                      .format(import_location))
    confirm = confirm.strip().lower()
    if confirm in ['y', 'yes']:
        util.write_out("Deleting {0}".format(import_location))
        util.check_call(['/usr/bin/rm', '-rf', import_location])
    util.write_out("Please restart docker daemon for the changes to take effect")

def import_images(import_location):
    """
    Method for importing docker images from a filesystem directory.
    """
    subdir = import_location + '/images'
    images = os.listdir(subdir)
    for image in images:
        util.write_out("Importing image: {0}".format(image[:12]))
        with open(subdir + '/' + image) as f:
            util.check_call([util.default_docker(), "load"], stdin=f)

def import_containers(graph, import_location):
    """
    Method for importing docker containers from a filesystem directory.
    """
    subdir = import_location + '/containers'
    containers = os.listdir(subdir)
    for cnt in containers:
        cnt = cnt[8:] # strip off the "migrate-" prefix
        util.write_out("Importing container: {0}".format(cnt[:12]))
        util.check_call([ATOMIC_LIBEXEC + '/migrate.sh',
                               'import',
                               '--container-id=' + cnt,
                               '--graph=' + graph,
                               '--import-location=' + import_location])

def tar_extract(srcfile, destdir):
    util.check_call(['/usr/bin/tar', '--extract', '--gzip', '--selinux',
                           '--file', srcfile, '--directory', destdir])

def import_volumes(graph, import_location):
    """
    Method for importing docker volumes from a filesystem directory.
    """

    volfile = import_location + '/volumes/volumeData.tar.gz'
    if os.path.isfile(volfile):
        util.write_out("Importing volumes")
        tar_extract(srcfile=volfile,
                    destdir=graph + '/volumes')

    vfsfile = import_location + '/volumes/vfsData.tar.gz'
    if os.path.isfile(vfsfile) and os.path.isdir(graph + "/vfs"):
        tar_extract(srcfile=vfsfile,
                    destdir=graph + '/vfs')



import os
import sys
import json
import pipes
from .client import AtomicDocker
from .syscontainers import SystemContainers

import requests

from . import util
import re
from Atomic.backends._docker_errors import NoDockerDaemon, DockerObjectNotFound
from docker.errors import NotFound
from .discovery import RegistryInspect, RegistryInspectError

def find_repo_tag(d, Id, image_name):
    def image_in_repotags(image_name, repotags):
        if image_name in repotags:
            return image_name
        for repotag in repotags:
            if repotag.startswith("{}:".format(image_name)):
                return image_name
        return None

    if not find_repo_tag.images:
        find_repo_tag.images = d.images()
    for image in find_repo_tag.images:
        repo_tag = image_in_repotags(image_name, image['RepoTags'])
        if repo_tag is not None:
            return repo_tag
        if Id == image["Id"]:
            return image["RepoTags"][0]
    return ""
find_repo_tag.images = None

class Atomic(object):
    results = '/var/lib/atomic'
    skull = (u"\u2620").encode('utf-8')

    def __init__(self):
        self.d = AtomicDocker()
        self.args = None
        self.command = None
        self.name = None
        self.image = None
        self.spc = False
        self.system = False
        self.setvalues = None
        self.inspect = None
        self.backend = None
        self.user = None
        self.force = False
        self._images = []
        self.containers = False
        self.images_cache = []
        self.images_all_cache = []
        self.active_containers = []
        self.docker_cmd = None
        self.debug = False
        self.is_python2 = (int(sys.version[0])) < 3
        self.useTTY = True
        self.syscontainers = SystemContainers()
        self.run_opts = None
        self.atomic_config = util.get_atomic_config()
        self.local_tokens = {}
        util.set_proxy()

    def __enter__(self):
        return self

    def __exit__(self, typ, value, traceback):
        try:
            self.d.close()
        except NoDockerDaemon:
            pass

    def docker_binary(self):
        if not self.docker_cmd:
            self.docker_cmd = util.default_docker()
        return self.docker_cmd

    def get_label(self, label, image=None):
        inspect = self._inspect_image(image)
        cfg = inspect.get("Config", None)
        if cfg:
            labels = cfg.get("Labels", [])
            if labels and label in labels:
                return labels[label]
        return ""

    def force_delete_containers(self):
        if self._inspect_image():
            image = self.image
            if self.image.find(":") == -1:
                image += ":latest"
            for c in self.get_containers():
                if c["Image"] == image:
                    self.d.remove_container(c["Id"], force=True)

    def pull(self):
        prevstatus = ""
        for line in self.d.pull(self.image, stream=True):
            bar = json.loads(line)
            status = bar['status']
            if prevstatus != status:
                util.write_out(status, "")
            if 'id' not in bar:
                continue
            if status == "Downloading":
                util.write_out(bar['progress'] + " ")
            elif status == "Extracting":
                util.write_out("Extracting: " + bar['id'])
            elif status == "Pull complete":
                pass
            elif status.startswith("Pulling"):
                util.write_out("Pulling: " + bar['id'])

            prevstatus = status
        util.write_out("")

    def set_args(self, args):
        self.args = args
        try:
            self.image = args.image
        except (NameError, AttributeError):
            pass
        try:
            self.command = args.command
        except (NameError, AttributeError):
            pass

        try:
            self.spc = args.spc
        except (NameError, AttributeError):
            pass

        try:
            self.system = args.system
        except (NameError, AttributeError):
            pass

        try:
            self.name = args.name
        except (NameError, AttributeError):
            pass

        try:
            self.force = args.force
        except (NameError, AttributeError):
            pass

        try:
            self.user = args.user
        except (NameError, AttributeError):
            pass

        if not self.name and self.image is not None:
            self.name = self.image.split("/")[-1].split(":")[0]
            if self.spc:
                self.name = self.name + "-spc"
            if self.system or self.user:
                self.name = self.syscontainers.get_default_system_name(self.image)

        try:
            if not self.name and args.container:
                self.name = args.container
        except (NameError, AttributeError):
            pass

        self.syscontainers.set_args(self.args)

    def _getconfig(self, key, default=None):
        assert self.inspect is not None
        cfg = self.inspect.get("Config")
        if cfg is None:
            return default
        val = cfg.get(key, default)
        if val is None:
            return default
        return val

    def _get_cmd(self):
        return self._getconfig("Cmd", ["/bin/sh"])

    def _get_labels(self):
        return self._getconfig("Labels", [])

    def _inspect_image(self, image=None):
        image = image or self.image
        try:
            if self.syscontainers.has_image(image):
                return self.syscontainers.inspect_system_image(image)
            return self.d.inspect_image(image)
        except (NotFound, requests.exceptions.ConnectionError):
            pass
        return None

    def _inspect_container(self, name=None):
        if name is None:
            name = self.name
        try:
            return self.d.inspect_container(name)
        except (NotFound, requests.exceptions.ConnectionError):
            pass
        return None

    def _get_args(self, label):
        labels = self._get_labels()
        for l in [label, label.lower(), label.capitalize(), label.upper()]:
            if l in labels:
                return labels[l].split()
        return None

    #def ps -> Atomic/ps.py

    #def run -> Atomic/run.py

    def quote(self, args):
        return list(map(pipes.quote, args))

    def cmd_env(self):
        newenv = dict(os.environ)
        newenv['NAME'] = self.name or ""
        newenv['IMAGE'] = self.image or ""

        if hasattr(self.args, 'opt1') and self.args.opt1:
            newenv['OPT1'] = os.path.expandvars(self.args.opt1)

        if hasattr(self.args, 'opt2') and self.args.opt2:
            newenv['OPT2'] = os.path.expandvars(self.args.opt2)

        if hasattr(self.args, 'opt3') and self.args.opt3:
            newenv['OPT3'] = os.path.expandvars(self.args.opt3)

        if not hasattr(self.args, 'PWD'):
            newenv['PWD'] = os.getcwd()

        default_uid = "0"
        with open("/proc/self/loginuid") as f:
            val = f.readline()
            if int(val) <= 2147483647:
                default_uid = val

        if "SUDO_UID" not in newenv:
            newenv["SUDO_UID"] = default_uid

        if 'SUDO_GID' not in newenv:
            newenv["SUDO_GID"] = default_uid

        if self.run_opts is not None:
            newenv["RUN_OPTS"] = self.run_opts
        return newenv

    def gen_cmd(self, cargs):
        args = []
        for c in cargs:
            if c == "IMAGE":
                args.append(self.image)
                continue
            if c == "IMAGE=IMAGE":
                args.append("IMAGE=%s" % self.image)
                continue
            if c == "NAME=NAME":
                args.append("NAME=%s" % self.name)
                continue
            if c == "NAME":
                args.append(self.name)
                continue
            args.append(c)
        if self.is_python2:
            return " ".join([x.decode('utf-8') for x in args])
        else:
            return " ".join(args)

    def get_fq_name(self, image_info):
        if not image_info['RepoTags']:
            return None

        if len(image_info['RepoTags']) > 1:
            if self.image in image_info['RepoTags']:
                return self.image

            possibles = []
            for i in image_info['RepoTags']:
                try:
                    possibles.append(self.get_fq_image_name(i))
                except RegistryInspectError:
                    possibles.append(None)

            if all(x==possibles[0] for x in possibles):
                return possibles[0]

            raise ValueError("\n{} is tagged with multiple repositories. "
                             "Try adding a tag to your input.\n".format(self.image))

        return image_info['RepoTags'][0]

    def is_iid(self):
        for i in self.get_images():
            if i['Id'].startswith(self.image):
                return True
        return False

    def _no_such_image(self):
        raise ValueError("Could not find any image matching '{}'"
                         .format(self.args.image))

    def is_dangling(self, image):
        if image == "<none>":
            return True
        return False

    def _container_exists(self, name):
        try:
            return self.syscontainers.get_checkout(name) or self._inspect_container(name)
        except ValueError:
            return None

    def help(self):
        if os.path.exists("/usr/bin/rpm-ostree"):
            return _('Atomic Management Tool')
        else:
            return _('Atomic Container Tool')

    def _get_layer(self, image):
        def get_label(label):
            return self.get_label(label, image["Id"])
        image = self._inspect_image(image)
        if not image:
            raise ValueError("Image '%s' does not exist" % self.image)
        version = ("%s-%s-%s" % (get_label("Name"), get_label("Version"),
                                 get_label("Release"))).strip("-")
        if 'Parent' in image:
            parent = image['Parent']
        else:
            parent = ""
        return({"Id": image['Id'], "Name": get_label("Name"),
                "Version": version, "RepoTags": image['RepoTags'],
                "Parent": parent})

    def get_layers(self):
        layers = []
        layer = self._get_layer(self.image)
        layers.append(layer)
        while layer["Parent"] != "":
            layer = self._get_layer(layer["Parent"])
            layers.append(layer)
        return layers

    def display(self, cmd):
        util.write_out(cmd)

    def sub_env_strings(self, in_string):
        """
        Performs substitutions on an input string based on defined
        environment variables.
        :param in_string: string to perform the subs on
        :return: string
        """
        # Perform variable subs
        in_string = util.expandvars(in_string, environ=self.cmd_env())

        # Replace undefined variables with blank
        in_string = re.sub(r'\$\{?\S*\}?', '', in_string)

        # Solve whitespacing
        in_string = " ".join(in_string.split())

        return in_string

    def ping(self):
        '''
        Check if the docker daemon is running; if not, exit with
        message and return code 1
        '''
        try:
            self.d.ping()
        except requests.exceptions.ConnectionError:
            raise NoDockerDaemon()

    def _is_container(self, identifier, active=False):
        '''
        Checks is the identifier is a container ID or container name.  If
        it is, returns the full container ID. Else it will return an
        AtomicError.  Takes optional keyword active, which signifies
        that you want to only deal with active containers.
        '''
        if active:
            active_cons = self.get_active_containers()
            active_con_ids = [x['Id'] for x in active_cons]
            cons = active_cons
        else:
            cons = self.get_containers()

        # First check if the container exists by whatever
        # identifier was given
        self.inspect = self._inspect_container(name=identifier)
        if self.inspect is not None:
            # Inspect found a match
            if not active:
                return self.inspect['Id']
            else:
                # Check if the container is active
                if self.inspect['Id'] in active_con_ids:
                    return self.inspect['Id']

        err_append = "Refine your search to narrow results."

        # The identifier might be a partial name?
        con_ids = []
        for con in cons:
            for name in con['Names']:
                if name.startswith("/{0}".format(identifier)):
                    con_ids.append(con['Id'])
                    break

        # More than one match was found
        if len(con_ids) > 1:
            raise AtomicError("Multiple matches were found for {0}. {1}"
                              .format(identifier, err_append))
        # No matches were found
        elif len(con_ids) < 1:
            active_err = '' if not active else 'active '
            error_msg = "Unable to find {0}container '{1}'".format(active_err,
                                                                   identifier)
            raise AtomicError(error_msg)
        else:
            self.inspect = self._inspect_container(con_ids[0])
            return con_ids[0]

    def _is_image(self, identifier):
        '''
        Checks is the identifier is a image ID or a matches an image name.
        If it finds a match, it returns the full image ID. Else it will
        return an AtomicError.
        '''
        err_append = "Refine your search to narrow results."
        image_info = self.get_images()

        inspect = self._inspect_image(image=identifier)
        if inspect is not None:
            self.inspect = inspect
            return inspect['Id']

        name_search = util.image_by_name(identifier, images=image_info)
        if len(name_search) > 0:
            if len(name_search) > 1:
                tmp_image = dict((x['Id'], x['RepoTags']) for x in image_info)
                repo_tags = []
                for name in name_search:
                    for repo_tag in tmp_image.get(name['Id']):
                        if repo_tag.find(identifier) > -1:
                            repo_tags.append(repo_tag)
                raise ValueError("Found more than one image possibly "
                                 "matching '{0}'. They are:\n    {1} \n{2}"
                                 .format(identifier, "\n    ".join(repo_tags),
                                         err_append))
            return name_search[0]['Id']
        # No dice
        raise AtomicError

    def is_duplicate_image(self, image):
        try:
            if self.syscontainers.has_image(image) and self.d.inspect_image(image):
                return True

            return False
        except (NotFound, requests.exceptions.ConnectionError):
            pass
        return False

    def get_input_id(self, identifier):
        '''
        Determine if the input "identifier" is valid.  Return the container or
        image ID when true and raise a ValueError when not
        '''
        try:
            return self._is_image(identifier)
        except AtomicError:
            pass
        try:
            return self._is_container(identifier)
        except AtomicError:
            pass

        if self.syscontainers.has_image(identifier):
            return identifier

        if self.syscontainers.get_checkout(identifier):
            return identifier

        raise DockerObjectNotFound(identifier)

    def _get_docker_images(self, get_all=False):
        try:
            images = self.d.images(all=get_all)
            for i in images:
                i["ImageType"] = "Docker"
                i["ImageId"] = i["Id"]
        except requests.exceptions.ConnectionError:
            raise NoDockerDaemon()
        return images

    def get_images(self, get_all=False):
        '''
        Wrapper function that should be used instead of querying docker
        multiple times for a list of images.
        '''
        if get_all:
            if len(self.images_all_cache) == 0:
                self.images_all_cache = self._get_docker_images(get_all=True) + self.syscontainers.get_system_images(get_all=True)
            return self.images_all_cache
        else:
            if len(self.images_cache) == 0:
                self.images_cache = self._get_docker_images() + self.syscontainers.get_system_images()
            return self.images_cache

    def get_containers(self):
        '''
        Wrapper function that should be used instead of querying docker
        multiple times for a list of containers
        '''
        if not self.containers:
            self.containers = self.d.containers(all=True)

        return self.containers + self.syscontainers.get_containers()

    def get_active_containers(self, refresh=False):
        '''
        Wrapper function for obtaining active containers.  Should be used
        instead of direct queries to docker
        '''
        if len(self.active_containers) == 0 or refresh:
            self.active_containers = self.d.containers(all=False)

        return self.active_containers

    def set_debug(self):
        if self.args.debug:
            self.debug = True

    def get_all_vulnerable_info(self):
        """
        Read and parse the /var/lib/atomic/scan_summary.json object.
        """
        try:
            return json.loads(open(os.path.join(self.results, "scan_summary.json"), "r").read())
        except (IOError, ValueError):
            return {}


    def get_vulnerable_ids(self):
        """
        Reads in /var/lib/atomic/scan_summary.json and returns a list of all
        the uuids that are vulnerable
        :return:
        """
        try:
            summary_results = json.loads(open(os.path.join(self.results, "scan_summary.json"), "r").read())
            vuln_ids = []

            for uuid in summary_results.keys():
                if 'Vulnerable' in summary_results[uuid]:
                    if summary_results[uuid]['Vulnerable']:
                        vuln_ids.append(uuid)
            return vuln_ids
        except IOError:
            return []

    def get_local_tokens(self):
        if len(self.local_tokens) == 0:
            self.local_tokens = self.load_local_tokens()
        return self.local_tokens

    @staticmethod
    def load_local_tokens():
        tokens = {}
        token_file_name = os.path.expanduser('~/.docker/config.json')
        if not os.path.exists(token_file_name):
            return {}
        with open(token_file_name) as token_file:
            token_data = json.load(token_file)
        try:
            for registry in token_data['auths']:
                reg_key = registry
                if registry == 'https://index.docker.io/v1/':
                    reg_key = 'docker.io'
                tokens[reg_key] = token_data['auths'][registry]['auth']
        except KeyError:
            # Just return a blank dict
            pass
        return tokens

    def get_fq_image_name(self, input_image):
        registry, repo, image, tag, digest = util.Decompose(input_image).all
        if not image:
            raise ValueError('Error parsing input: "{}" invalid'.format(input_image))
        if all([True if x else False for x in [registry, image, tag]]):
            img = registry
            if repo:
                img += "/{}".format(repo)
            img += "/{}:{}".format(image, tag)
            return img
        if not registry:
            ri = RegistryInspect(registry, repo, image, tag, digest=digest, debug=self.args.debug, orig_input=self.image)
            return ri.find_image_on_registry()


class AtomicError(Exception):
    pass




from . import util


class RegistryInspectError(Exception):
    pass


class RegistryAuthError(Exception):
    pass


class RegistryInspect():

    def __init__(self, registry=None, repo=None, image=None, tag=None, digest=None, orig_input=None, debug=False):
        self.debug = debug
        self.registries = util.get_registries()
        self.registry = registry
        self.repo = repo
        self.image = image
        self.tag = tag
        self.digest = digest
        self.orig_input = orig_input
        self._remote_inspect = None
        self._fqdn = None

        if self.debug:
            util.output_json(self.registries)

    @property
    def fqdn(self):
        if not self._fqdn:
            self._fqdn = self.assemble_fqdn(include_tag=True) if self.registry else self.find_image_on_registry()
        return self._fqdn

    @fqdn.setter
    def fqdn(self, value):
        self._fqdn = value

    def inspect(self):
        if self.registry:
            inspect_data = util.skopeo_inspect("docker://{}".format(self.fqdn), return_json=True)
        else:
            inspect_data = self._remote_inspect
        inspect_data['Tag'] = self.tag
        inspect_data['Name'] = self.assemble_fqdn(include_tag=False)
        return inspect_data

    def get_manifest(self, return_json=True):
        assert(self.fqdn is not None)
        return util.skopeo_inspect("docker://{}".format(self.fqdn), return_json=return_json, args=['--raw'])

    @property
    def remote_id(self):
        result = self.get_manifest()
        if result.get('config'):
            return result['config'].get('digest', None)
        return None

    def assemble_fqdn(self, include_tag=True, registry=None):
        fqdn = "{}".format(registry or self.registry)
        if self.repo:
            fqdn = "{}/{}".format(fqdn, self.repo)
        elif fqdn == "docker.io": # and no repo specified
            fqdn = fqdn + "/library"
        fqdn += "/{}".format(self.image)
        if include_tag:
            if self.tag:
                fqdn += ":{}".format(self.tag)
            elif self.digest:
                fqdn += "@{}".format(self.digest)
        return fqdn

    def find_image_on_registry(self, quiet=False):
        """
        Find the fully qualified image name for given input when
        registry is unknown
        :return: String fqdn
        """
        if self.debug:
            for i in [x for x in self.registries if x['search']]:
                util.write_out(repr(i))

        registries = [i['name'] for i in [x for x in self.registries if x['search']]]
        for registry in registries:
            fqdn = self.assemble_fqdn(registry=registry, include_tag=True)
            if not quiet:
                util.write_out("Trying {}...".format(fqdn))
            try:
                result = util.skopeo_inspect("docker://{}".format(fqdn), return_json=True)
                self._remote_inspect = result
                return fqdn
            except ValueError as e:
                if not quiet:
                    util.write_err("Failed: {}".format(e))
                continue
        raise RegistryInspectError("Unable to find {}".format(self.orig_input))

